import pytest
import tensorflow_probability.substrates.jax as tfp
from jax import numpy as jnp

from jaxns.framework.model import Model
from jaxns.framework.prior import Prior

tfpd = tfp.distributions


def test_gh95():
    def prior_model():
        x = yield Prior(dist_or_value=jnp.asarray(0.))
        return x

    def log_likelihood(x):
        return jnp.sum(x)

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
    assert model.U_ndims == 0


def test_parametrised_singular():
    def prior_model():
        x = yield Prior(dist_or_value=jnp.asarray(0.), name='x').parametrised()
        return x

    def log_likelihood(x):
        return jnp.sum(x)

    with pytest.raises(ValueError, match="Cannot parametrise a prior without distribution."):
        _ = Model(prior_model=prior_model, log_likelihood=log_likelihood)


def test_parametrised():
    def prior_model():
        x = yield Prior(tfpd.Uniform(),name='x').parametrised()
        return x

    def log_likelihood(x):
        return jnp.sum(x)

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
    assert model.num_params == 1
    assert model.U_ndims == 0

    log_prob_joint = model.log_prob_joint(model.U_placeholder, allow_nan=True)
    assert log_prob_joint.shape == ()
    assert log_prob_joint == 0.5, "Didn't init at median of uniform"



from typing import Optional

from jax import random, numpy as jnp
from jax.scipy import special

from jaxns.internals.log_semiring import cumulative_logsumexp
from jaxns.internals.types import FloatArray, IntArray, PRNGKey

__all__ = ['random_ortho_matrix',
           'resample_indicies']


def random_ortho_matrix(key, n, special_orthogonal: bool = False):
    """
    Samples a random orthonormal n by n matrix from Stiefels manifold.
    From https://stackoverflow.com/a/38430739

    Args:
        key: PRNG seed
        n: Size of matrix, draws from O(num_options) group.

    Returns: random [num_options,num_options] matrix with determinant = +-1
    """
    H = random.normal(key, shape=(n, n))
    Q, R = jnp.linalg.qr(H)
    if special_orthogonal:
        R *= jnp.sign(R)
    Q = Q @ jnp.diag(jnp.sign(jnp.diag(R)))
    return Q


def resample_indicies(key: PRNGKey, log_weights: Optional[FloatArray] = None, S: Optional[int] = None,
                      replace: bool = False, num_total: Optional[int] = None) -> IntArray:
    """
    Get resample indicies according to a given weighting, with or without replacement.

    Args:
        key: PRNGKey
        log_weights: Optional log weights
        S: Optional number of samples. Computes effective sample size from log weights if not given.
        replace: whether to use replacement or not.
        num_total: Optional total sample size to use, must be given if `replace=False` and `log_weights=None`

    Returns:
        index array given the take indicies to resample at.
    """
    if S is None:
        if log_weights is None:
            raise ValueError("Need log_weights if S is not given.")
        # ESS = (sum w)^2 / sum w^2
        S = int(jnp.exp(2. * special.logsumexp(log_weights) - special.logsumexp(2. * log_weights)))

    if replace:
        if log_weights is not None:
            # use cumulative_logsumexp because some log_weights could be really small
            log_p_cuml = cumulative_logsumexp(log_weights)
            log_r = log_p_cuml[-1] + jnp.log(1. - random.uniform(key, (S,)))
            idx = jnp.searchsorted(log_p_cuml, log_r)
        else:
            if num_total is None:
                raise ValueError("Need num_total if log_weights is None.")
            log_p_cuml = jnp.log(jnp.arange(num_total))
            log_r = log_p_cuml[-1] + jnp.log(1. - random.uniform(key, (S,)))
            idx = jnp.searchsorted(log_p_cuml, log_r)
    else:
        if log_weights is not None:
            g = -random.gumbel(key, shape=log_weights.shape) - log_weights
        else:
            if num_total is None:
                raise ValueError("Need num_total if log_weights is None.")
            g = -random.gumbel(key, shape=(num_total,))
        idx = jnp.argsort(g)[:S]
    return idx



from typing import NamedTuple, Optional, Union, Any, Callable, Tuple, Dict, List, TypeVar

import chex
import numpy as np
from jax import numpy as jnp

__all__ = [
    'EvidenceCalculation',
    'TerminationCondition',
    'StaticStandardNestedSamplerState',
    'NestedSamplerResults',
    'PRNGKey',
    'IntArray',
    'FloatArray',
    'BoolArray',
    'float_type',
    'int_type',
    'complex_type',
    'LikelihoodType',
    'UType',
    'XType',
    'LikelihoodInputType',
    'RandomVariableType',
    'MeasureType'
]

float_type = jnp.result_type(float)
int_type = jnp.result_type(int)
complex_type = jnp.result_type(complex)

PRNGKey = chex.PRNGKey
FloatArray = chex.Array
IntArray = chex.Array
BoolArray = chex.Array

LikelihoodType = Callable[..., FloatArray]
RandomVariableType = TypeVar('RandomVariableType')
MeasureType = TypeVar('MeasureType')
LikelihoodInputType = Tuple[RandomVariableType, ...]  # Likelihood conditional variables
UType = FloatArray  # Sample space type
XType = Dict[str, RandomVariableType]  # Prior variable type


class EvidenceCalculation(NamedTuple):
    """
    Contains a running estimate of evidence and related quantities.
    """
    log_L: FloatArray
    log_X_mean: FloatArray
    log_X2_mean: FloatArray
    log_Z_mean: FloatArray
    log_ZX_mean: FloatArray
    log_Z2_mean: FloatArray
    log_dZ_mean: FloatArray
    log_dZ2_mean: FloatArray


class TerminationCondition(NamedTuple):
    """
    Contains the termination conditions for the nested sampling run.

    Args:
        ess: The effective sample size, if the ESS (Kish's estimate) is greater than this the run will terminate.
        evidence_uncert: The uncertainty in the evidence, if the uncertainty is less than this the run will terminate.
        live_evidence_frac: Depreceated use dlogZ.
        dlogZ: Terminate if log(Z_current + Z_remaining) - log(Z_current) < dlogZ. Default log(1 + 1e-2)
        max_samples: Terminate if the number of samples exceeds this.
        max_num_likelihood_evaluations: Terminate if the number of likelihood evaluations exceeds this.
        log_L_contour: Terminate if this log(L) contour is reached. A contour is reached if any dead point
            has log(L) > log_L_contour. Uncollected live points are not considered.
        efficiency_threshold: Terminate if the efficiency (num_samples / num_likelihood_evaluations) is less than this,
            for the last shrinkage iteration.
    """
    ess: Optional[FloatArray] = jnp.asarray(jnp.inf, float_type)
    evidence_uncert: Optional[FloatArray] = jnp.asarray(0., float_type)
    live_evidence_frac: Optional[FloatArray] = None # Depreceated use dlogZ
    dlogZ: Optional[FloatArray] = jnp.asarray(np.log(1. + 1e-3), float_type) #
    max_samples: Optional[IntArray] = jnp.asarray(jnp.iinfo(int_type).max, int_type)
    max_num_likelihood_evaluations: Optional[IntArray] = jnp.asarray(jnp.iinfo(int_type).max, int_type)
    log_L_contour: Optional[FloatArray] = jnp.asarray(jnp.inf, float_type)
    efficiency_threshold: Optional[FloatArray] = jnp.asarray(0., float_type)

    def __and__(self, other):
        return TerminationConditionConjunction(conds=[self, other])

    def __or__(self, other):
        return TerminationConditionDisjunction(conds=[self, other])


class TerminationConditionConjunction(NamedTuple):
    """
    A conjunction of termination conditions, e.g. term_cond1 & term_cond2
    """
    conds: List[Union['TerminationConditionDisjunction', 'TerminationConditionConjunction', TerminationCondition]]


class TerminationConditionDisjunction(NamedTuple):
    """
    A disjunction of termination conditions, e.g. term_cond1 | term_cond2
    """
    conds: List[Union['TerminationConditionDisjunction', TerminationConditionConjunction, TerminationCondition]]


class NestedSamplerResults(NamedTuple):
    """
    Results of the nested sampling run.
    """
    log_Z_mean: FloatArray  # estimate of E[log(Z)]
    log_Z_uncert: FloatArray  # estimate of StdDev[log(Z)]
    ESS: FloatArray  # estimate of Kish's effective sample size
    H_mean: FloatArray  # estimate of E[int log(L) L dp/Z]
    samples: XType  # Dict of arrays with leading dimension num_samples
    parametrised_samples: XType  # Dict of arrays with leading dimension num_samples
    U_samples: UType  # Dict of arrays with leading dimension num_samples
    log_L_samples: FloatArray  # log(L) of each sample
    log_dp_mean: FloatArray  # log(E[dZ]) of each sample, where dZ is how much it contributes to the total evidence.
    log_X_mean: FloatArray  # log(E[U]) of each sample
    log_posterior_density: FloatArray  # log(P( theta | D )) log posteriori density
    num_live_points_per_sample: IntArray  # how many live points were taken for the samples.
    num_likelihood_evaluations_per_sample: IntArray  # how many likelihood evaluations were made per sample.
    total_num_samples: IntArray  # int, the total number of samples collected.
    total_phantom_samples: IntArray  # int, the total number of phantom samples collected.
    total_num_likelihood_evaluations: IntArray  # how many likelihood evaluations were made in total
    log_efficiency: FloatArray  # log(total_num_samples / total_num_likelihood_evaluations)
    termination_reason: IntArray  # this will be an int reflecting the reason for termination


class SignedLog(NamedTuple):
    """
    Represents a signed value in log-space
    """
    log_abs_val: jnp.ndarray
    sign: Union[jnp.ndarray, Any]


class Sample(NamedTuple):
    U_sample: UType  # [..., D] sample in U-space
    log_L_constraint: FloatArray  # [...,] log(L) constraint
    log_L: FloatArray  # [...,] log(L) of sample
    num_likelihood_evaluations: IntArray  # [...,] number of likelihood evaluations


class StaticStandardSampleCollection(NamedTuple):
    sender_node_idx: IntArray  # [N] with values in [0, N]
    log_L: MeasureType  # [N] log(L) of each sample
    U_samples: UType  # [N, D] samples in U-space
    num_likelihood_evaluations: IntArray  # [N] number of likelihood evaluations for each sample
    phantom: BoolArray  # [N] whether the sample is a phantom sample

class TerminationRegister(NamedTuple):
    num_samples_used: IntArray
    evidence_calc: EvidenceCalculation
    evidence_calc_with_remaining: EvidenceCalculation
    num_likelihood_evaluations: IntArray
    log_L_contour: FloatArray
    efficiency: FloatArray
    plateau: BoolArray

class StaticStandardNestedSamplerState(NamedTuple):
    key: PRNGKey
    next_sample_idx: IntArray  # the next sample insert index <==> the number of samples
    sample_collection: StaticStandardSampleCollection
    front_idx: IntArray  # the index of the front of the live points within sample collection



from typing import Tuple, Optional, NamedTuple

import jax.numpy as jnp

from jaxns.internals.cumulative_ops import cumulative_op_static, cumulative_op_dynamic
from jaxns.internals.log_semiring import LogSpace
from jaxns.internals.tree_structure import SampleTreeGraph, count_crossed_edges
from jaxns.internals.types import MeasureType, EvidenceCalculation, float_type, IntArray, FloatArray


def compute_enclosed_prior_volume(sample_tree: SampleTreeGraph) -> MeasureType:
    """
    Compute the enclosed prior volume of the likelihood constraint.

    Args:
        sample_tree: The sample tree graph.

    Returns:
        The log enclosed prior volume.
    """
    live_point_counts = count_crossed_edges(sample_tree=sample_tree)

    def op(log_X, num_live_points):
        X_mean = LogSpace(log_X)
        # T_mean = LogSpace(jnp.log(num_live_points) - jnp.log(num_live_points + 1.))
        # T_mean = LogSpace(jnp.log(1.) - jnp.log(1. + 1./num_live_points))
        T_mean = LogSpace(- jnp.logaddexp(0., -jnp.log(num_live_points)))
        next_X_mean = X_mean * T_mean
        return next_X_mean.log_abs_val

    _, log_X = cumulative_op_static(op=op, init=jnp.asarray(-jnp.inf, float_type),
                                    xs=live_point_counts.num_live_points)
    return log_X


class EvidenceUpdateVariables(NamedTuple):
    num_live_points: FloatArray
    log_L_next: FloatArray


def _update_evidence_calc_op(carry: EvidenceCalculation, y: EvidenceUpdateVariables) -> EvidenceCalculation:
    # num_live_points = num_live_points.astype(float_type)
    next_L = LogSpace(y.log_L_next)
    L_contour = LogSpace(carry.log_L)
    midL = LogSpace(jnp.log(0.5)) * (next_L + L_contour)
    X_mean = LogSpace(carry.log_X_mean)
    X2_mean = LogSpace(carry.log_X2_mean)
    Z_mean = LogSpace(carry.log_Z_mean)
    ZX_mean = LogSpace(carry.log_ZX_mean)
    Z2_mean = LogSpace(carry.log_Z2_mean)
    dZ2_mean = LogSpace(carry.log_dZ2_mean)

    # T_mean = LogSpace(jnp.log(num_live_points) - jnp.log(num_live_points + 1.))
    # T_mean = LogSpace(jnp.log(1.) - jnp.log(1. + 1./num_live_points))
    T_mean = LogSpace(- jnp.logaddexp(0., -jnp.log(y.num_live_points)))
    # T_mean = LogSpace(- jnp.logaddexp(0., -jnp.log(num_live_points)))
    t_mean = LogSpace(- jnp.log(y.num_live_points + 1.))
    # T2_mean = LogSpace(jnp.log(num_live_points) - jnp.log( num_live_points + 2.))
    # T2_mean = LogSpace(jnp.log(1.) - jnp.log(1. + 2./num_live_points))
    T2_mean = LogSpace(- jnp.logaddexp((0.), jnp.log(2.) - jnp.log(y.num_live_points)))
    # T2_mean = LogSpace(- jnp.logaddexp(jnp.log(2.), -jnp.log(num_live_points)))
    t2_mean = LogSpace(jnp.log(2.) - jnp.log(y.num_live_points + 1.) - jnp.log(y.num_live_points + 2.))
    # tT_mean = LogSpace(jnp.log(num_live_points) - jnp.log(num_live_points + 1.) - jnp.log(num_live_points + 2.))
    # tT_mean = LogSpace(jnp.log(1.) - jnp.log(1. + 1./num_live_points) - jnp.log(num_live_points + 2.))
    tT_mean = LogSpace(- jnp.logaddexp(0., -jnp.log(y.num_live_points)) - jnp.log(y.num_live_points + 2.))
    # tT_mean = LogSpace(- jnp.logaddexp(0., -jnp.log(num_live_points)) - jnp.log(num_live_points + 2.))

    dZ_mean = X_mean * t_mean * midL
    next_X_mean = X_mean * T_mean
    next_X2_mean = X2_mean * T2_mean
    next_Z_mean = Z_mean + dZ_mean
    next_ZX_mean = ZX_mean * T_mean + X2_mean * tT_mean * midL
    next_Z2_mean = Z2_mean + LogSpace(jnp.log(2.)) * ZX_mean * t_mean * midL + (X2_mean * t2_mean * midL ** 2)
    next_dZ2_mean = dZ2_mean + (X2_mean * t2_mean * midL ** 2)

    next_evidence_calculation = EvidenceCalculation(
        log_L=y.log_L_next.astype(float_type),
        log_X_mean=next_X_mean.log_abs_val.astype(float_type),
        log_X2_mean=next_X2_mean.log_abs_val.astype(float_type),
        log_Z_mean=next_Z_mean.log_abs_val.astype(float_type),
        log_Z2_mean=next_Z2_mean.log_abs_val.astype(float_type),
        log_ZX_mean=next_ZX_mean.log_abs_val.astype(float_type),
        log_dZ_mean=dZ_mean.log_abs_val.astype(float_type),
        log_dZ2_mean=next_dZ2_mean.log_abs_val.astype(float_type)
    )

    return next_evidence_calculation


def update_evicence_calculation(evidence_calculation: EvidenceCalculation,
                                update: EvidenceUpdateVariables) -> EvidenceCalculation:
    """
    Update the evidence statistics with a new sample.

    Args:
        evidence_calculation: The current evidence statistics.
        y: The update variables.

    Returns:
        The updated evidence statistics.
    """
    return _update_evidence_calc_op(evidence_calculation, update)


def init_evidence_calc() -> EvidenceCalculation:
    """
    Initialise the evidence statistics.

    Returns:
        The initial evidence statistics.
    """
    return EvidenceCalculation(
        log_L=jnp.asarray(-jnp.inf, float_type),
        log_X_mean=jnp.asarray(0., float_type),
        log_X2_mean=jnp.asarray(0., float_type),
        log_Z_mean=jnp.asarray(-jnp.inf, float_type),
        log_ZX_mean=jnp.asarray(-jnp.inf, float_type),
        log_Z2_mean=jnp.asarray(-jnp.inf, float_type),
        log_dZ_mean=jnp.asarray(-jnp.inf, float_type),
        log_dZ2_mean=jnp.asarray(-jnp.inf, float_type)
    )


def compute_evidence_stats(log_L: MeasureType, num_live_points: FloatArray, num_samples: Optional[IntArray] = None) -> \
        Tuple[EvidenceCalculation, EvidenceCalculation]:
    """
    Compute the evidence statistics along the shrinkage process.

    Args:
        log_L: The log likelihoods of the samples.
        num_live_points: The number of live points at each sample.
        num_samples: The number of samples to use. If None, all samples are used.

    Returns:
        The final evidence statistics, and the evidence statistics for each sample.
    """
    init = init_evidence_calc()
    xs = EvidenceUpdateVariables(
        num_live_points=num_live_points.astype(float_type),
        log_L_next=log_L
    )
    if num_samples is not None:
        stop_idx = num_samples
        final_accumulate, result = cumulative_op_dynamic(op=_update_evidence_calc_op, init=init, xs=xs, stop_idx=stop_idx)
    else:
        final_accumulate, result = cumulative_op_static(op=_update_evidence_calc_op, init=init, xs=xs)
    final_evidence_calculation = final_accumulate
    per_sample_evidence_calculation = result
    return final_evidence_calculation, per_sample_evidence_calculation



import logging
from typing import Tuple, Optional, Union

import haiku as hk
import jax.nn
import tensorflow_probability.substrates.jax as tfp
from jax import numpy as jnp

from jaxns.framework.bases import BaseAbstractPrior, BaseAbstractDistribution
from jaxns.framework.distribution import Distribution
from jaxns.internals.types import FloatArray, IntArray, BoolArray, XType, UType, float_type

tfpd = tfp.distributions

__all__ = [
    "Prior",
    "InvalidPriorName"
]

logger = logging.getLogger('jaxns')


class InvalidPriorName(Exception):
    """
    Raised when a prior name is already taken.
    """

    def __init__(self, name: Optional[str] = None):
        super(InvalidPriorName, self).__init__(f'Prior name {name} already taken by another prior.')


class SingularPrior(BaseAbstractPrior):
    """
    Represents a singular prior, which has no inverse transformation, but does have a log_prob
        (at the singular value).
    """

    def __init__(self, value: jnp.ndarray, dist: BaseAbstractDistribution, name: Optional[str] = None):
        super().__init__(name=name)
        self.value = value
        self.dist = dist

    def __repr__(self):
        return f"{self.value} -> {self.dist}"

    def _dtype(self):
        return self.dist.dtype

    def _base_shape(self) -> Tuple[int, ...]:
        return (0,)  # Singular prior has no base shape

    def _shape(self) -> Tuple[int, ...]:
        return self.dist.shape

    def _forward(self, U: UType) -> Union[FloatArray, IntArray, BoolArray]:
        return self.value

    def _inverse(self, X: XType) -> FloatArray:
        return jnp.asarray([], float_type)

    def _log_prob(self, X: XType) -> FloatArray:
        return self.dist.log_prob(X)


class Prior(BaseAbstractPrior):
    """
    Represents a generative prior.
    """

    def __init__(self, dist_or_value: Union[tfpd.Distribution, BaseAbstractDistribution, jnp.ndarray],
                 name: Optional[str] = None):
        super(Prior, self).__init__(name=name)
        if isinstance(dist_or_value, tfpd.Distribution):
            self._type = 'dist'
            self._dist = Distribution(dist_or_value)
        elif isinstance(dist_or_value, BaseAbstractDistribution):
            self._type = 'dist'
            self._dist = dist_or_value
        else:
            self._type = 'value'
            self._value = jnp.asarray(dist_or_value)
        self.name = name

    @property
    def dist(self) -> BaseAbstractDistribution:
        if self._type != 'dist':
            raise ValueError(f"Wrong type, got {self._type}")
        return self._dist

    @property
    def value(self) -> jnp.ndarray:
        if self._type != 'value':
            raise ValueError(f"Wrong type, got {self._type}")
        return self._value

    def _base_shape(self) -> Tuple[int, ...]:
        if self._type == 'value':
            return (0,)
        elif self._type == 'dist':
            return self.dist.base_shape
        else:
            raise NotImplementedError()

    def _shape(self) -> Tuple[int, ...]:
        if self._type == 'value':
            return self.value.shape
        elif self._type == 'dist':
            return self.dist.shape
        else:
            raise NotImplementedError()

    def _dtype(self):
        if self._type == 'value':
            return self.value.dtype
        elif self._type == 'dist':
            return self.dist.dtype
        else:
            raise NotImplementedError()

    def _forward(self, U: UType) -> Union[FloatArray, IntArray, BoolArray]:
        if self._type == 'value':
            return self.value
        elif self._type == 'dist':
            return self.dist.forward(U)
        else:
            raise NotImplementedError()

    def _inverse(self, X: XType) -> FloatArray:
        if self._type == 'value':
            return jnp.asarray([], float_type)
        elif self._type == 'dist':
            return self.dist.inverse(X)
        else:
            raise NotImplementedError()

    def _log_prob(self, X: XType) -> FloatArray:
        if self._type == 'value':
            return jnp.asarray(0., float_type)
        elif self._type == 'dist':
            return self.dist.log_prob(X=X)
        else:
            raise NotImplementedError()

    def parametrised(self) -> SingularPrior:
        """
        Convert this prior into a non-Bayesian parameter, that takes a single value in the model, but still has an associated
        log_prob. The parameter is registered as a `hk.Parameter` with added `_param` name suffix.

        Returns:
            A singular prior.
        """
        if self._type == 'value':
            raise ValueError("Cannot parametrise a prior without distribution.")
        return prior_to_parametrised_singular(self)


def prior_to_parametrised_singular(prior: Prior) -> SingularPrior:
    """
    Convert a prior into a non-Bayesian parameter, that takes a single value in the model, but still has an associated
    log_prob. The parameter is registered as a `hk.Parameter` with added `_param` name suffix.

    To constrain the parameter we use a Normal parameter with centre on unit cube, and scale covering the whole cube,
    as the base representation. This base representation covers the whole real line and be reliably used with SGD, etc.

    Args:
        prior: any prior

    Returns:
        A parameter representing the prior.
    """
    if prior.name is None:
        raise ValueError("Prior must have a name to be parametrised.")
    name = f"{prior.name}_param"
    # Initialises at median of distribution.
    init_value = jnp.zeros(prior.base_shape, dtype=float_type)
    if init_value.size == 0:
        logger.warning(f"Creating a zero-sized parameter for {prior.name}. Probably unintended.")
    norm_U_base_param = hk.get_parameter(
        name=name,
        shape=prior.base_shape,
        dtype=float_type,
        init=hk.initializers.Constant(init_value)
    )
    # transform [-inf, inf] -> [0,1]
    # Sigmoid is faster than ndtr to save FLOPs
    # U_base_param = ndtr(norm_U_base_param)
    U_base_param = jax.nn.sigmoid(norm_U_base_param)
    param = prior.forward(U_base_param)
    return SingularPrior(value=param, dist=prior.dist, name=prior.name)



from typing import NamedTuple, Tuple, Literal

import numpy as np
import pylab as plt
from jax import numpy as jnp, vmap, random, tree_map, lax
from jax._src.scipy.special import gammaln

from jaxns.internals.log_semiring import LogSpace
from jaxns.samplers.multi_ellipsoid.em_gmm import em_gmm
from jaxns.internals.types import IntArray, FloatArray, PRNGKey, BoolArray
from jaxns.internals.types import UType, int_type, float_type

__all__ = [
    'ellipsoid_clustering',
    'sample_multi_ellipsoid',
    'MultEllipsoidState',
]


class EllipsoidParams(NamedTuple):
    mu: FloatArray  # [K, D] Ellipsoids centres
    radii: FloatArray  # [K, D] Ellsipoids radii
    rotation: FloatArray  # [K, D, D] Ellipsoids rotation matrices


class MultEllipsoidState(NamedTuple):
    params: EllipsoidParams
    cluster_id: IntArray  # [N] the cluster index of each point


def log_ellipsoid_volume(radii):
    D = radii.shape[0]
    return jnp.log(2.) - jnp.log(D) + 0.5 * D * jnp.log(jnp.pi) - gammaln(0.5 * D) + jnp.sum(jnp.log(radii))


def bounding_ellipsoid(points: UType, mask: FloatArray) -> Tuple[FloatArray, FloatArray]:
    """
    Use empirical mean and covariance as approximation to bounding ellipse.

    Args:
        points: [N, D] points to fit ellipsoids to
        mask: [N] mask of which points to consider

    Returns:
        mu, cov
    """
    mu = jnp.average(points, weights=mask, axis=0)
    dx = points - mu
    cov = jnp.average(dx[:, :, None] * dx[:, None, :], weights=mask, axis=0)
    return mu, cov


def covariance_to_rotational(cov: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """
    (x - mu)^T inv(cov) (x - mu) = (x - mu)^T J @ J.T (x - mu)

    where J.T is composed of un-rotation and un-scaling:

    J.T = diag(1/radii) @ rotation.T <==> J = rotation @ diag(1/radii)

    Now since, cov = U @ diag(s) @ V.H we have

    J @ J.T = inv(U @ diag(s) @ V.H) = V @ diag(1/s) @ U.H

    ==> J.T = diag(1/sqrt(s)) @ U.H
    ==> radii = sqrt(s), rotation = U

    Args:
        cov:

    Returns:
        radii, rotation
    """
    u, s, vh = jnp.linalg.svd(cov)
    radii_min = jnp.finfo(s.dtype).eps
    radii = jnp.maximum(jnp.sqrt(s), radii_min)
    rotation = u
    return radii, rotation


def ellipsoid_params(points: UType, mask: FloatArray) -> EllipsoidParams:
    """
    If the ellipsoid is defined by

    (x - mu)^T C (x - mu) = 1

    where C = L @ L.T and L = diag(1/radii) @ rotation.T

    then this returns the mu, radius and rotation matrices of the ellipsoid.

    Args:
        points: [N, D] points to fit ellipsoids to
        mask: [N] mask of which points to consider

    Returns:
        mu [D], radii [D] rotation [D,D]
    """
    # get ellipsoid mean and covariance
    mu, Sigma = bounding_ellipsoid(points=points, mask=mask)
    radii, rotation = covariance_to_rotational(Sigma)

    # Compute scale factor for radii to enclose all points.
    # for all i (points[i] - mu) @ inv(Sigma) / scale**2 @ (points[i] - mu) <= 1
    # for all i (points[i] - mu) @ (L @ L.T) @ (points[i] - mu) <= scale**2
    rho = vmap(lambda x: maha_ellipsoid(x=x, mu=mu, radii=radii, rotation=rotation))(points)

    rho_max = jnp.max(jnp.where(mask, rho, 0.))
    radii *= jnp.sqrt(rho_max)

    return EllipsoidParams(mu=mu, radii=radii, rotation=rotation)


def ellipsoid_to_circle(point: FloatArray, mu: FloatArray, radii: FloatArray, rotation: FloatArray) -> FloatArray:
    """
    Apply a linear map that would turn an ellipsoid into a sphere.
    Args:
        point: [D] point to transform
        mu: [D] center of ellipse
        radii: [D] radii of ellipse
        rotation: [D,D] rotation matrix of ellipse

    Returns:
        a transformed point of shape [D]
    """
    return jnp.diag(jnp.reciprocal(radii)) @ rotation.T @ (point - mu)


def circle_to_ellipsoid(point: FloatArray, mu: FloatArray, radii: FloatArray, rotation: FloatArray) -> FloatArray:
    """
    Apple a linear map that would turn a sphere into an ellipsoid

    Args:
        point: [D] point to transform
        mu: [D] center of ellipse
        radii: [D] radii of ellipse
        rotation: [D,D] rotation matrix of ellipse

    Returns:
        a transformed point of shape [D]
    """
    return mu + (rotation @ jnp.diag(radii) @ point)


def maha_ellipsoid(x: FloatArray, mu: FloatArray, radii: FloatArray, rotation: FloatArray) -> FloatArray:
    """
    Compute the Mahalanobis distance.

    Args:
        x: point [D]
        mu: center of ellipse [D]
        radii: radii of ellipse [D]
        rotation: rotation matrix [D, D]

    Returns:
        The Mahalanobis distance of `x` to `mu`.
    """
    u_circ = ellipsoid_to_circle(x, mu, radii, rotation)
    return u_circ @ u_circ


def point_in_ellipsoid(x: FloatArray, mu: FloatArray, radii: FloatArray, rotation: FloatArray) -> BoolArray:
    """
    Determine if a given point is inside a closed ellipse.

    Args:
        x: point [D]
        mu: center of ellipse [D]
        radii: radii of ellipse [D]
        rotation: rotation matrix [D, D]

    Returns:
        True iff x is inside the closed ellipse
    """
    return jnp.less_equal(maha_ellipsoid(x, mu, radii, rotation), jnp.asarray(1., x.dtype))


def sample_ellipsoid(key: PRNGKey, mu: FloatArray, radii: FloatArray, rotation: FloatArray,
                     unit_cube_constraint: bool = False) -> FloatArray:
    """
    Sample uniformly inside an ellipsoid. When unit_cube_constraint=True then reject points outside unit-cube.

    Args:
        key:
        mu: [D]
        radii: [D]
        rotation: [D,D]
        unit_cube_constraint: whether to restrict to the closed unit-cube.

    Returns:
        i.i.d. sample from ellipsoid of shape [D]
    """

    def _single_sample(key):
        direction_key, radii_key = random.split(key, 2)
        direction = random.normal(direction_key, shape=radii.shape)
        direction = direction / jnp.linalg.norm(direction)
        t = random.uniform(radii_key) ** (1. / radii.size)
        u_circ = direction * t
        R = rotation * radii
        u = R @ u_circ + mu
        return u

    def body(state):
        (key, _, _) = state
        key, sample_key = random.split(key, 2)
        u = _single_sample(sample_key)
        done = jnp.all((u <= 1) & (u >= 0))
        return (key, done, u)

    if unit_cube_constraint:
        (_, _, u) = lax.while_loop(lambda s: ~s[1],
                                   body,
                                   (key, jnp.asarray(False), mu))
    else:
        u = _single_sample(key)
    return u


def compute_depth_ellipsoids(point: FloatArray, mu: FloatArray, radii: FloatArray, rotation: FloatArray,
                             constraint_unit_cube: bool = False) -> IntArray:
    """
    Compute overlap of ellipsoids at point. Points outside the domain are given infinite depth.

    Args:
        point: [D] point to compute depth at.
        mu: [K, D] means of ellispoids
        radii: [K, D] radii of ellipsoids
        rotation: [K, D, D] rotation matrices of ellipsoids
        constraint_unit_cube: bool, whether domain is clipped to closed unit-cube.

    Returns:
        scalar representing overlap of ellipsoids.
    """
    # in any of the ellipsoids
    contained_in = vmap(lambda mu, radii, rotation: point_in_ellipsoid(point, mu, radii, rotation))(mu, radii, rotation)
    depth = jnp.sum(contained_in)
    if constraint_unit_cube:
        # outside cube
        outside_unit_cube = jnp.any(point < 0.) | jnp.any(point > 1.)
        depth = jnp.where(outside_unit_cube, jnp.iinfo(depth.dtype).max, depth)
    return depth


def sample_multi_ellipsoid(key: PRNGKey, mu: FloatArray, radii: FloatArray, rotation: FloatArray,
                           unit_cube_constraint: bool = True) -> Tuple[IntArray, FloatArray]:
    """
    Sample from a set of intersecting ellipsoids.
    When unit_cube_constraint=True then reject points outside the closed unit-cube.

    Args:
        key: PRNGKey
        mu: [K, D] centres of ellipses
        radii: [K, D] radii of ellipses
        rotation: [K,D,D] rotation matrices of ellipses

    Returns:
        ellipsoid selected, and a sample point i.i.d. sampled from union of ellipsoids, of shape [D]
    """
    # u(t) = R @ (x + t * num_options) + c
    #     u(t) == 1
    #     1-c = R@x + t * R@num_options
    #     t = ((1 - c) - R@x)/R@num_options
    K, D = radii.shape
    log_VE = vmap(log_ellipsoid_volume)(radii)
    log_p = log_VE  # - logsumexp(log_VE)

    def body(state):
        (i, _, key, done, _) = state
        key, accept_key, sample_key, select_key = random.split(key, 4)
        k = random.categorical(select_key, log_p)
        mu_k = mu[k, :]
        radii_k = radii[k, :]
        rotation_k = rotation[k, :, :]
        u_test = sample_ellipsoid(sample_key, mu_k, radii_k, rotation_k, unit_cube_constraint=False)
        depth = compute_depth_ellipsoids(u_test, mu, radii, rotation, constraint_unit_cube=unit_cube_constraint)
        done = random.uniform(accept_key) < jnp.reciprocal(depth)
        return (i + 1, k, key, done, u_test)

    _, k, _, _, u_accept = lax.while_loop(lambda state: ~state[3],
                                          body,
                                          (jnp.array(0), jnp.array(0), key, jnp.array(False), jnp.zeros(D)))
    return k, u_accept


def log_coverage_scale(log_VE, log_VS, D):
    """
    Computes the required scaling relation such that
    V(E) = max(V(E), V(S))
    where the scaling is to be applied to each radius.
    Args:
        log_VE:
        log_VS:
        D:
    Returns:
    """
    return jnp.maximum(0., (log_VS - log_VE) / D)


class ClusterSplitResult(NamedTuple):
    unsorted_cluster_id: IntArray  # unsorted cluster id, using 0/1 to indicate the child.
    log_VS0: FloatArray
    params0: EllipsoidParams
    log_VS1: FloatArray
    params1: EllipsoidParams
    successful_split: BoolArray


def _multinest_split(key: PRNGKey, params: EllipsoidParams, points: FloatArray, mask: BoolArray, log_VS: FloatArray,
                     em_init: bool = False,
                     patience: int = 1):
    """
    Use's Multinest's method to partition points

    Args:
        key: PRNGKey
        params: ellipsoid params of points that are being split (same as used for log VE)
        points: [N, D] points to split
        mask: [N] mask only those points which should be split
        log_VS: estimate of logV(S) of the set of points
        em_init: whether to use kmeans to initialise the clustering
        patience: how long to wait before seeing improvement

    Returns:
        cluster_id, log_VS0, params0, log_VS1, params1
    """
    init_key, volume_key = random.split(key, 2)

    N, D = points.shape
    n_S = jnp.sum(mask)
    # calculate bounding ellipsoid

    ###
    # input is essentially log_VS
    if em_init:
        # do Euclidean kmean clustering
        cluster_id, (_, _, _), _ = em_gmm(
            key=init_key,
            data=points,
            mask=mask,
            n_components=2,
            n_iters=100
        )
    else:
        # Split the ellipsoid in half
        j_max = jnp.argmax(params.radii)
        n = jnp.where(jnp.arange(params.radii.size) == j_max,
                      jnp.asarray(1., float_type),
                      jnp.asarray(0., float_type)
                      )
        p = params.rotation @ (jnp.diag(params.radii) @ n)
        q = points - params.mu
        proj = q @ p
        cluster_id = jnp.where(proj >= jnp.asarray(0., float_type),
                               jnp.asarray(0, int_type),
                               jnp.asarray(1, int_type))
        # # assign to random clusters: child0 or child1
        # cluster_id = random.randint(init_key, shape=(N,), minval=0, maxval=2)

    class CarryState(NamedTuple):
        iter: IntArray
        done: BoolArray
        cluster_id: IntArray
        log_VS0: FloatArray
        params0: EllipsoidParams
        log_VS1: FloatArray
        params1: EllipsoidParams
        min_loss: FloatArray
        iters_no_improvement: IntArray

    def body(body_state: CarryState):
        mask0 = mask & (body_state.cluster_id == 0)
        mask1 = mask & (body_state.cluster_id == 1)
        # estimate volumes of current clustering
        n0 = jnp.sum(mask0)
        n1 = jnp.sum(mask1)
        log_VS0 = log_VS + jnp.log(n0) - jnp.log(n_S)
        log_VS1 = log_VS + jnp.log(n1) - jnp.log(n_S)
        # construct E_1, E_2 and compute volumes
        params0 = ellipsoid_params(points=points, mask=mask0)
        log_VE0 = log_ellipsoid_volume(params0.radii)
        params1 = ellipsoid_params(points=points, mask=mask1)
        log_VE1 = log_ellipsoid_volume(params1.radii)
        # enlarge to at least cover V(S1) and V(S2)
        log_scale0 = log_coverage_scale(log_VE0, log_VS0, D)
        log_scale1 = log_coverage_scale(log_VE1, log_VS1, D)
        radii0 = jnp.exp(jnp.log(params0.radii) + log_scale0)
        radii1 = jnp.exp(jnp.log(params1.radii) + log_scale1)
        log_VE0 = log_VE0 + log_scale0 * D
        log_VE1 = log_VE1 + log_scale1 * D
        params0 = params0._replace(radii=radii0)
        params1 = params1._replace(radii=radii1)

        # compute reassignment metrics
        maha0 = vmap(lambda point: maha_ellipsoid(point,
                                                  mu=params0.mu,
                                                  radii=params0.radii,
                                                  rotation=params0.radii))(points)
        maha1 = vmap(lambda point: maha_ellipsoid(point, mu=params1.mu,
                                                  radii=params1.radii,
                                                  rotation=params1.radii))(points)
        h0 = LogSpace(log_VE0) * LogSpace(jnp.log(maha0)) / LogSpace(log_VS0)
        h1 = LogSpace(log_VE1) * LogSpace(jnp.log(maha1)) / LogSpace(log_VS1)
        # reassign biggest violator
        abs_delta_F = (h0 - h1).abs()  # N
        masked_log_abs_delta_F = jnp.where(mask, abs_delta_F.log_abs_val, -jnp.inf)
        reassign_idx = jnp.argmax(masked_log_abs_delta_F)
        new_id = jnp.where(masked_log_abs_delta_F[reassign_idx] > -jnp.inf,
                           jnp.asarray(1, int_type) - cluster_id[reassign_idx],
                           cluster_id[reassign_idx])
        new_cluster_id = cluster_id.at[reassign_idx].set(new_id)

        # new_cluster_id = jnp.where(mask & (h0.log_abs_val < h1.log_abs_val),
        #                            jnp.asarray(0., int_type),
        #                            jnp.asarray(1., int_type))

        # new_cluster_k = jnp.where(log_h1 < log_h2, 0, 1)
        log_V_sum = jnp.logaddexp(log_VE0, log_VE1)
        new_loss = log_V_sum - log_VS  # If scaling happened, then this will be zero
        loss_decreased = new_loss < body_state.min_loss
        iters_no_improvement = jnp.where(loss_decreased, 0, body_state.iters_no_improvement + 1)
        min_loss = jnp.where(loss_decreased, new_loss, body_state.min_loss)
        ###
        # i / delay / loss_decreased / new_loss / min_loss
        # 0 / 0 / True / a / a
        # 1 / 1 / False / b / a
        # 2 / 2 / False / a / a
        # 3 / 3 / False / b / a
        # 4 / 4 / False / a / a
        cluster_mapping_unchanged = jnp.all((new_cluster_id == body_state.cluster_id) | jnp.bitwise_not(mask))
        done = cluster_mapping_unchanged \
               | (iters_no_improvement >= patience) \
               | (n0 < D + 1) \
               | (n1 < D + 1) \
               | jnp.isnan(log_V_sum)

        return CarryState(
            iter=body_state.iter + jnp.asarray(1, int_type),
            done=done,
            cluster_id=new_cluster_id,
            log_VS0=log_VS0,
            params0=params0,
            log_VS1=log_VS1,
            params1=params1,
            min_loss=min_loss,
            iters_no_improvement=iters_no_improvement
        )

    # Done to start with if not at least D+1 points per ellipsoid possible
    done = (n_S < 2 * (D + 1))

    init_state = CarryState(
        iter=jnp.array(0),
        done=done,
        cluster_id=cluster_id,
        log_VS0=jnp.array(-jnp.inf),
        params0=EllipsoidParams(mu=jnp.zeros(D), radii=jnp.zeros(D), rotation=jnp.eye(D)),
        log_VS1=jnp.array(-jnp.inf),
        params1=EllipsoidParams(mu=jnp.zeros(D), radii=jnp.zeros(D), rotation=jnp.eye(D)),
        min_loss=jnp.asarray(jnp.inf),
        iters_no_improvement=jnp.asarray(0, int_type)
    )

    output_state: CarryState = lax.while_loop(lambda state: ~state.done,
                                              body,
                                              init_state)
    return output_state.cluster_id, output_state.log_VS0, output_state.params0, output_state.log_VS1, output_state.params1


def _em_gmm_split(key: PRNGKey, points: FloatArray, mask: BoolArray, log_VS: FloatArray):
    """
    Use's EM Gaussian mixture model to partition points.

    Args:
        key: PRNGKey
        points: [N, D] points to split
        mask: [N] mask only those points which should be split
        log_VS: estimate of logV(S) of the set of points

    Returns:
        cluster_id, log_VS0, params0, log_VS1, params1
    """
    N, D = points.shape
    n_S = jnp.sum(mask)

    # do Euclidean kmean clustering
    cluster_id, (_, _, _), _ = em_gmm(
        key=key,
        data=points,
        mask=mask,
        n_components=2,
        n_iters=100
    )

    mask0 = mask & (cluster_id == 0)
    mask1 = mask & (cluster_id == 1)
    # estimate volumes of current clustering
    n0 = jnp.sum(mask0)
    n1 = jnp.sum(mask1)
    log_VS0 = log_VS + jnp.log(n0) - jnp.log(n_S)
    log_VS1 = log_VS + jnp.log(n1) - jnp.log(n_S)
    # construct E_1, E_2 and compute volumes
    params0 = ellipsoid_params(points=points, mask=mask0)
    log_VE0 = log_ellipsoid_volume(params0.radii)
    params1 = ellipsoid_params(points=points, mask=mask1)
    log_VE1 = log_ellipsoid_volume(params1.radii)
    # enlarge to at least cover V(S1) and V(S2)
    log_scale0 = log_coverage_scale(log_VE0, log_VS0, D)
    log_scale1 = log_coverage_scale(log_VE1, log_VS1, D)
    radii0 = jnp.exp(jnp.log(params0.radii) + log_scale0)
    radii1 = jnp.exp(jnp.log(params1.radii) + log_scale1)
    params0 = params0._replace(radii=radii0)
    params1 = params1._replace(radii=radii1)

    return cluster_id, log_VS0, params0, log_VS1, params1


def cluster_split(key: PRNGKey, params: EllipsoidParams, points: FloatArray, mask: BoolArray, log_VS: FloatArray,
                  method: Literal['multinest', 'em_gmm']) -> ClusterSplitResult:
    """
    Splits a set of points into two ellipsoids such that the enclosed volume is as close to V(S) without being less.
    V(S) should be an estimate of the true volume contained by the points.

    Args:
        key: PRNGKey
        params: ellipsoid params of points that are being split (same as used for log VE)
        points: [N, D] points to split
        mask: [N] mask only those points which should be split
        log_VS: estimate of logV(S) of the set of points
        method: what method to use for splitting. Available are: 'multinest','em_gmm'

    Returns:
        cluster split results
    """

    N, D = points.shape
    # calculate bounding ellipsoid
    # volume of ellipsoid, already have E scaled so that V(E) >= V(S)
    log_VE = log_ellipsoid_volume(params.radii)

    # We always have
    if method == 'em_gmm':
        cluster_id, log_VS0, params0, log_VS1, params1 = _em_gmm_split(key=key, points=points, mask=mask,
                                                                       log_VS=log_VS)
    elif method == 'multinest':
        cluster_id, log_VS0, params0, log_VS1, params1 = _multinest_split(key=key, params=params, points=points,
                                                                          mask=mask,
                                                                          log_VS=log_VS,
                                                                          em_init=False,
                                                                          patience=1)
    else:
        raise ValueError(f"Invalid method {method}")

    # Imperfect sampling condition
    # (0)       V(A) <= V(S1), and V(B) <= V(S2)

    # V(A) <= V(S1) = V(S) V(A) / (V(A) + V(B))
    # V(A) (V(A) + V(B)) <= V(S) V(A)
    # V(A)^2 + V(A)V(B) - V(S) V(A) <= 0
    # V(B)^2 + V(A)V(B) - V(S) V(B) <= 0
    # V(A)^2 - V(B)^2 - V(S) (V(A) - V(B)) <= 0
    # (V(A) + V(B)) (V(A) - V(B)) - V(S) (V(A) - V(B)) <= 0
    # (V(A) + V(B)) - V(S) <= 0, (V(A) - V(B)) != 0
    #
    # Bounding ellipsoid condition
    # (1)       V(S1) <= V(E_A), and V(S2) <= V(E_B)
    # (1.1)     from (0) => V(A) + V(B) <= V(S1) + V(S2) <= V(E_A) + V(E_B)

    # Disjoint partitioning condition
    # (2)       V(S1) + V(S2) = V(S)
    # (2.1)     from (1.1) => V(A) + V(B) <= V(S1) + V(S2) = V(S) <= V(E_A) + V(E_B)
    # (2.2)     from (2) => V(S1 ^ S2) = 0

    # Disjoint partitioning condition
    # (3)       V(A) + V(B) = V(A v B)
    # (3.1)     from (2) => V(A ^ B) = 0

    # Bounding ellipsoid condition
    # (4)       V(S) <= V(E)
    # (4.1)     with (2.1) => V(A) + V(B) <= V(S1) + V(S2) = V(S) <= V(E)

    # (5)       Good split <=> V(S1) ~ V(E_A), and V(S2) ~ V(E_B)
    # (5.1)     from (4.1) Good split <=> V(A) + V(B) <= V(S1) + V(S2) = V(S) <~ V(E_B) + V(E_B)  <= V(E)

    # (6)       Good sampling <=> V(A) ~ V(S1), and V(B) ~ V(S2)
    # (6.1)     from (5.1) Good sampling <=> V(A) + V(B) <~ V(S1) + V(S2) = V(S) <= V(E)

    # (7)       Good split and Good sampling <=> V(A) + V(B) <~ V(S1) + V(S2) = V(S) <~ V(E_B) + V(E_B) <= V(E)

    # We take the condition for success:
    #  V(E_B) + V(E_B) is closer to V(S) than V(E) (from 5.1)

    mask0 = mask & (cluster_id == 0)
    mask1 = mask & (cluster_id == 1)

    log_VE_A = log_ellipsoid_volume(params0.radii)
    log_VE_B = log_ellipsoid_volume(params1.radii)

    V_sum = LogSpace(log_VE_A) + LogSpace(log_VE_B)

    good_split = (V_sum.log_abs_val < log_VE)

    successful_split = good_split \
                       & jnp.bitwise_not(jnp.isnan(log_VE_A)) \
                       & jnp.bitwise_not(jnp.isnan(log_VE_A)) \
                       & (jnp.sum(mask0) >= (D + 1)) \
                       & (jnp.sum(mask1) >= (D + 1))

    return ClusterSplitResult(unsorted_cluster_id=cluster_id, log_VS0=log_VS0,
                              params0=params0, log_VS1=log_VS1,
                              params1=params1, successful_split=successful_split)


def plot_ellipses(params: EllipsoidParams, show: bool = True):
    """
    Plots ellipses.

    Args:
        params: ellipsoid parameters to plot
        show: whether to show figure
    """
    theta = jnp.linspace(0., 2 * jnp.pi, 100)
    circle = jnp.stack([jnp.cos(theta), jnp.sin(theta)], axis=1)
    for mu, radii, rotation in zip(params.mu, params.radii, params.rotation):
        ellipse = vmap(lambda point: circle_to_ellipsoid(point, mu, radii, rotation))(circle)
        plt.plot(ellipse[:, 0], ellipse[:, 1], c=np.random.uniform(size=3))
    if show:
        plt.show()


def ellipsoid_clustering(key: PRNGKey, points: FloatArray, log_VS: FloatArray,
                         max_num_ellipsoids: int,
                         method: Literal['multinest', 'em_gmm'] = 'em_gmm') -> MultEllipsoidState:
    """
    Partition live_points into 2^depth ellipsoids in depth-first order.

    Args:
        key:PRNGKey
        points: [N, D] points to partition
        log_VS: expected true volume of points
        max_num_ellipsoids: the maximum number of ellipsoids

    Returns:
        params of multi-ellipsoids and cluster id of points
    """

    N, D = points.shape
    if max_num_ellipsoids < 1:
        raise ValueError(f"max_num_ellipsoids should be >= 1, got {max_num_ellipsoids}.")
    K = max_num_ellipsoids

    # Construct the initial state
    init_ellipsoid = ellipsoid_params(points=points, mask=jnp.ones(N, jnp.bool_))

    log_VE = log_ellipsoid_volume(init_ellipsoid.radii)
    log_scale = log_coverage_scale(log_VE, log_VS, D)
    radii = jnp.exp(jnp.log(init_ellipsoid.radii) + log_scale)
    # log_VE = log_VE + log_scale * D
    init_ellipsoid = init_ellipsoid._replace(radii=radii)

    # state is zeros except first ellipsoid
    cluster_id = jnp.zeros(N, dtype=int_type)
    params = EllipsoidParams(
        mu=jnp.zeros((K, D), float_type),
        radii=jnp.zeros((K, D), float_type),
        rotation=jnp.zeros((K, D, D), float_type)
    )
    params: EllipsoidParams = tree_map(lambda x, y: x.at[0].set(y), params, init_ellipsoid)
    state = MultEllipsoidState(
        cluster_id=cluster_id,
        params=params
    )

    # Initial tracking parameters
    log_VS_subclusters = jnp.asarray([log_VS] + [-jnp.inf] * (K - 1))
    done_splitting = jnp.isneginf(log_VS_subclusters)
    split_depth = jnp.zeros([K], int_type)

    # TODO: compare performance with scan

    class CarryType(NamedTuple):
        key: PRNGKey
        next_k: IntArray
        state: MultEllipsoidState
        done_splitting: BoolArray
        split_depth: IntArray
        log_VS_subclusters: FloatArray

    def body(body_state: CarryType) -> CarryType:
        key, split_key = random.split(body_state.key, 2)
        # Select the depth we work on now: bread first selection ==> min depth first (excluding done splits)
        select_split = jnp.argmin(
            jnp.where(body_state.done_splitting, jnp.iinfo(body_state.split_depth.dtype).max, body_state.split_depth)
        )
        mask = body_state.state.cluster_id == select_split
        # estimated volume in sub-cluster
        log_VS = body_state.log_VS_subclusters[select_split]

        # params of ellipsoid
        params = tree_map(lambda x: x[select_split], body_state.state.params)

        # Perform a split on points in the given mask
        # Strategy: if no split we replace child0 with parent and child1 gets zero-size ellipsoid that has no members.
        cluster_split_result: ClusterSplitResult = cluster_split(
            key=split_key,
            params=params,
            points=points,
            mask=mask,
            log_VS=log_VS,
            method=method
        )

        # Update the parameters in given component that is being split with child 0
        params = tree_map(lambda x, y: jnp.where(cluster_split_result.successful_split, x.at[select_split].set(y), x),
                          body_state.state.params,
                          cluster_split_result.params0)
        # Update the parameters in `next_k` with child 1
        params = tree_map(
            lambda x, y: jnp.where(cluster_split_result.successful_split, x.at[body_state.next_k].set(y), x),
            params,
            cluster_split_result.params1)
        # select_split stays the same cluster_id taking on child 0, but next_k gets child 1
        cluster_id = jnp.where(
            cluster_split_result.successful_split & (cluster_split_result.unsorted_cluster_id == 1) & mask,
            body_state.next_k,
            body_state.state.cluster_id)
        state = body_state.state._replace(params=params, cluster_id=cluster_id)

        # If success => next_k is not done, (and select_k is not done, as previously set)
        # Else select_k is done (next_k stays done, as previously set)
        done_splitting = jnp.where(cluster_split_result.successful_split,
                                   body_state.done_splitting.at[body_state.next_k].set(False),
                                   body_state.done_splitting.at[select_split].set(True))

        # If success => update split depth
        new_depth = body_state.split_depth[select_split] + jnp.asarray(1, int_type)
        split_depth = jnp.where(cluster_split_result.successful_split,
                                body_state.split_depth.at[select_split].set(new_depth),
                                body_state.split_depth)
        split_depth = jnp.where(cluster_split_result.successful_split,
                                split_depth.at[body_state.next_k].set(new_depth),
                                split_depth)

        # If success => update estimated subcluster volumes
        log_VS_subclusters = jnp.where(cluster_split_result.successful_split,
                                       body_state.log_VS_subclusters.at[select_split].set(cluster_split_result.log_VS0),
                                       body_state.log_VS_subclusters)
        log_VS_subclusters = jnp.where(cluster_split_result.successful_split,
                                       log_VS_subclusters.at[body_state.next_k].set(cluster_split_result.log_VS1),
                                       log_VS_subclusters)

        # TODO: (verify) I think next_k should only increment if successful split, as otherwise it uses up space.
        next_k = jnp.where(cluster_split_result.successful_split,
                           body_state.next_k + jnp.asarray(1, int_type),
                           body_state.next_k)

        return CarryType(
            key=key,
            next_k=next_k,
            state=state,
            done_splitting=done_splitting,
            split_depth=split_depth,
            log_VS_subclusters=log_VS_subclusters
        )

    def cond(body_state: CarryType) -> BoolArray:
        done = jnp.all(body_state.done_splitting) | (body_state.next_k == K)
        return jnp.bitwise_not(done)

    init_body_state = CarryType(
        key=key,
        next_k=jnp.asarray(1, int_type),
        state=state,
        done_splitting=done_splitting,
        split_depth=split_depth,
        log_VS_subclusters=log_VS_subclusters
    )

    output_state = lax.while_loop(cond, body, init_body_state)

    return output_state.state



from timeit import default_timer

import jax
import numpy as np
from jax import tree_map, numpy as jnp, random

from jaxns.internals.tree_structure import SampleTreeGraph, SampleLivePointCounts, count_crossed_edges, \
    count_intervals_naive, \
    plot_tree, count_old, count_crossed_edges_less_fast, concatenate_sample_trees, unbatch_state
from jaxns.internals.types import StaticStandardNestedSamplerState, StaticStandardSampleCollection


def test_naive():
    S = SampleTreeGraph(
        sender_node_idx=jnp.asarray([0, 0, 0, 1, 2, 3]),
        log_L=jnp.asarray([1, 2, 3, 4, 5, 6])
    )
    plot_tree(S)
    expected = SampleLivePointCounts(
        samples_indices=jnp.asarray([0, 1, 2, 3, 4, 5]),
        num_live_points=jnp.asarray([3, 3, 3, 3, 2, 1])
    )
    assert all(tree_map(lambda x, y: np.array_equal(x, y), count_intervals_naive(S), expected))
    assert all(tree_map(lambda x, y: np.array_equal(x, y), count_intervals_naive(S), count_old(S)))


def test_basic():
    S = SampleTreeGraph(
        sender_node_idx=jnp.asarray([0, 0, 0, 1, 2, 3]),
        log_L=jnp.asarray([1, 2, 3, 4, 5, 6])
    )
    assert all(tree_map(lambda x, y: np.array_equal(x, y), count_crossed_edges(S), count_intervals_naive(S)))
    assert all(tree_map(lambda x, y: np.array_equal(x, y), count_crossed_edges(S), count_old(S)))
    assert all(tree_map(lambda x, y: np.array_equal(x, y), count_crossed_edges(S), count_crossed_edges_less_fast(S)))

    S = SampleTreeGraph(
        sender_node_idx=jnp.asarray([0, 0, 0, 1, 3, 2]),
        log_L=jnp.asarray([1, 2, 3, 4, 6, 5])
    )

    assert all(tree_map(lambda x, y: np.array_equal(x, y), count_crossed_edges(S), count_intervals_naive(S)))
    assert all(tree_map(lambda x, y: np.array_equal(x, y), count_crossed_edges(S), count_old(S)))
    assert all(tree_map(lambda x, y: np.array_equal(x, y), count_crossed_edges(S), count_crossed_edges_less_fast(S)))


def test_with_num_samples():
    S1 = SampleTreeGraph(
        sender_node_idx=jnp.asarray([0, 0, 0, 1, 2, 3, 4, 5, 0, 0]),
        log_L=jnp.asarray([1, 2, 3, 4, 5, 6, 7, 8, jnp.inf, jnp.inf])
    )
    num_samples = 8

    S2 = SampleTreeGraph(
        sender_node_idx=jnp.asarray([0, 0, 0, 1, 2, 3, 4, 5]),
        log_L=jnp.asarray([1, 2, 3, 4, 5, 6, 7, 8])
    )

    assert all(tree_map(lambda x, y: np.array_equal(x[:num_samples], y),
                        count_crossed_edges(S1, num_samples),
                        count_crossed_edges(S2)))

    output = count_crossed_edges(S1, num_samples)
    print(output)
    np.testing.assert_array_equal(output.num_live_points[num_samples:], 0)


def test_random_tree():
    np.random.seed(42)
    log_L = [0]
    parent_idx = []

    for idx in range(10):
        new_log_L = log_L[idx] + np.random.uniform(low=0, high=1 - log_L[idx]) ** 4
        log_L.append(new_log_L)
        parent_idx.append(idx)

    for idx in range(10):
        for _ in range(5):
            new_log_L = np.random.uniform(low=log_L[idx], high=1.)
            log_L.append(new_log_L)
            parent_idx.append(idx)

    S = SampleTreeGraph(
        sender_node_idx=jnp.asarray(parent_idx),
        log_L=jnp.asarray(log_L)[1:]
    )

    plot_tree(S)

    assert all(tree_map(lambda x, y: np.array_equal(x, y), count_crossed_edges(S), count_intervals_naive(S)))
    assert all(tree_map(lambda x, y: np.array_equal(x, y), count_crossed_edges(S), count_old(S)))
    assert all(tree_map(lambda x, y: np.array_equal(x, y), count_crossed_edges(S), count_crossed_edges_less_fast(S)))

    T = count_crossed_edges_less_fast(S)
    import pylab as plt
    plt.plot(S.log_L[T.samples_indices], T.num_live_points)
    plt.show()


def test_speed():
    np.random.seed(42)
    log_L = [0]
    parent_idx = []

    for _ in range(100000):
        idx = np.random.choice(len(log_L))
        new_log_L = log_L[idx] + np.random.uniform(low=0.5, high=1.)
        log_L.append(new_log_L)
        parent_idx.append(idx)

    S = SampleTreeGraph(
        sender_node_idx=jnp.asarray(parent_idx),
        log_L=jnp.asarray(log_L)[1:]
    )

    def time_method(f):
        f_jit = jax.jit(f)
        T = f_jit(S)
        T.samples_indices.block_until_ready()
        t0 = default_timer()
        for _ in range(100):
            T = f_jit(S)
            T.samples_indices.block_until_ready()
        dt = (default_timer() - t0) / 100
        print(f.__name__, dt)
        return dt

    # time_method(count_intervals_naive)
    time_method(count_crossed_edges)
    time_method(count_crossed_edges_less_fast)
    time_method(count_old)


def test_concatenate_sample_trees():
    S1 = SampleTreeGraph(
        sender_node_idx=jnp.asarray([0, 1]),
        log_L=jnp.asarray([1, 2])
    )
    expect = SampleTreeGraph(
        sender_node_idx=jnp.asarray([0, 1, 0, 3]),
        log_L=jnp.asarray([1, 2, 1, 2])
    )
    S = concatenate_sample_trees([S1, S1])
    assert all(S.sender_node_idx == expect.sender_node_idx)
    assert all(S.log_L == expect.log_L)

    # Three graphs
    S = concatenate_sample_trees([S1, S1, S1])
    expect = SampleTreeGraph(
        sender_node_idx=jnp.asarray([0, 1, 0, 3, 0, 5]),
        log_L=jnp.asarray([1, 2, 1, 2, 1, 2])
    )
    assert all(S.sender_node_idx == expect.sender_node_idx)
    assert all(S.log_L == expect.log_L)


def test_unbatch_state():
    # Example of two batched states, first dimension is has 1 sample, second has 2 samples. Max samples is 2 for both.
    # num_live_points=1 for both.
    single_state_1 = StaticStandardNestedSamplerState(
        key=random.PRNGKey(42),
        next_sample_idx=jnp.asarray(2),
        sample_collection=StaticStandardSampleCollection(
            sender_node_idx=jnp.asarray([0, 1, 0]),
            log_L=jnp.asarray([1., 2., jnp.inf]),
            U_samples=jnp.asarray([1., 1., 0.])[:, None],
            num_likelihood_evaluations=jnp.asarray([1, 2, 0]),
            phantom=jnp.asarray([True, False, False])
        ),
        front_idx=jnp.asarray([1])
    )

    single_state_2 = StaticStandardNestedSamplerState(
        key=random.PRNGKey(42),
        next_sample_idx=jnp.asarray(2),
        sample_collection=StaticStandardSampleCollection(
            sender_node_idx=jnp.asarray([0, 1, 0]),
            log_L=jnp.asarray([1.1, 2.1, jnp.inf]),
            U_samples=jnp.asarray([1., 1., 0.])[:, None],
            num_likelihood_evaluations=jnp.asarray([1, 2, 0]),
            phantom=jnp.asarray([True, False, False])
        ),
        front_idx=jnp.asarray([1])
    )

    batched_state = tree_map(lambda x, y: jnp.stack([x, y], axis=0), single_state_1, single_state_2)

    unbatched_state = unbatch_state(batched_state)

    print(unbatched_state)

    # The second element of first batch is not measure so it should be at end of unbatched
    expected_state = StaticStandardNestedSamplerState(
        key=random.PRNGKey(42),
        next_sample_idx=jnp.asarray(4),
        sample_collection=StaticStandardSampleCollection(
            sender_node_idx=jnp.asarray([0, 1, 0, 0, 4, 0]),
            log_L=jnp.asarray([1., 2., jnp.inf, 1.1, 2.1, jnp.inf]),
            U_samples=jnp.asarray([1., 1., 0., 1., 1., 0.])[:, None],
            num_likelihood_evaluations=jnp.asarray([1, 2, 0, 1, 2, 0]),
            phantom=jnp.asarray([True, False, False, True, False, False])
        ),
        front_idx=jnp.asarray([1, 4])
    )

    expected_state = StaticStandardNestedSamplerState(
        key=random.PRNGKey(42),
        next_sample_idx=jnp.asarray(4),
        sample_collection=StaticStandardSampleCollection(
            sender_node_idx=jnp.asarray([0, 0, 1, 2, 0, 0]),
            log_L=jnp.asarray([1., 1.1, 2., 2.1, jnp.inf, jnp.inf]),
            U_samples=jnp.asarray([1., 1., 1., 1., 0., 0.])[:, None],
            num_likelihood_evaluations=jnp.asarray([1, 1, 2, 2, 0, 0]),
            phantom=jnp.asarray([True, True, False, False, False, False])
        ),
        front_idx=jnp.asarray([2, 3])
    )

    # compare pytrees
    for a, b in zip(jax.tree_util.tree_leaves(unbatched_state), jax.tree_util.tree_leaves(expected_state)):
        assert jnp.all(a == b)



import logging
from typing import TypeVar, NamedTuple, Tuple, Optional

from jax import numpy as jnp, random, tree_map, lax

from jaxns.framework.bases import BaseAbstractModel
from jaxns.internals.cumulative_ops import cumulative_op_static
from jaxns.internals.types import PRNGKey, FloatArray, BoolArray, Sample, int_type, StaticStandardNestedSamplerState, \
    UType, \
    IntArray, float_type, StaticStandardSampleCollection
from jaxns.samplers.abc import SamplerState
from jaxns.samplers.bases import SeedPoint, BaseAbstractMarkovSampler

__all__ = [
    'MultiDimSliceSampler'
]

logger = logging.getLogger('jaxns')

T = TypeVar('T')


def _slice_bounds(key: PRNGKey, point_U0: FloatArray, num_restrict_dims: int) -> Tuple[FloatArray, FloatArray]:
    """
    Get the slice bounds, randomly selecting which dimensions to slice in.

    Args:
        key: PRNGKey
        point_U0: the seed point
        num_restrict_dims: the number of dimensions to slice in

    Returns:
        left, and right bounds of slice
    """
    if num_restrict_dims is not None:
        slice_dims = random.choice(key, point_U0.size, shape=(num_restrict_dims,), replace=False)
        left = point_U0.at[slice_dims].set(jnp.zeros(num_restrict_dims, point_U0.dtype))
        right = point_U0.at[slice_dims].set(jnp.ones(num_restrict_dims, point_U0.dtype))
    else:
        left = jnp.zeros_like(point_U0)
        right = jnp.ones_like(point_U0)
    return left, right


def _new_sample(key: PRNGKey, left: FloatArray, right: FloatArray) -> UType:
    return random.uniform(key=key, shape=left.shape, dtype=left.dtype, minval=left, maxval=right)


def _shrink_region(point_U: UType, point_U0: UType, left: FloatArray, right: FloatArray) -> Tuple[
    FloatArray, FloatArray]:
    """
    Shrink the region to the left and right of the point_U0.

    Args:
        point_U: the point to shrink to
        point_U0: the origin of the slice
        left: the left bound of the slice
        right: the right bound of the slice

    Returns:
        new left, and right bounds of slice
    """
    # if point_U is on the 'right' side then we shrink the 'right' side to it.
    # same of 'left'
    left = jnp.where(point_U < point_U0,
                     jnp.maximum(left, point_U),
                     left)
    right = jnp.where(point_U > point_U0,
                      jnp.minimum(right, point_U),
                      right)

    return left, right


def _new_proposal(key: PRNGKey, seed_point: SeedPoint, num_restrict_dims: int, log_L_constraint: FloatArray,
                  model: BaseAbstractModel) -> Tuple[FloatArray, FloatArray, IntArray]:
    """
    Sample from a slice about a seed point.

    Args:
        key: PRNG key
        seed_point: the seed point to sample from
        num_restrict_dims: how many dimensions to restrict the slice to
        log_L_constraint: the constraint to sample within
        model: the model to sample from

    Returns:
        point_U: the new sample
        log_L: the log-likelihood of the new sample
        num_likelihood_evaluations: the number of likelihood evaluations performed
    """

    class Carry(NamedTuple):
        key: PRNGKey
        left: FloatArray
        right: FloatArray
        point_U: UType
        log_L: FloatArray
        num_likelihood_evaluations: IntArray

    def cond(carry: Carry) -> BoolArray:
        close_to_zero_interval = jnp.all((carry.right - carry.left) <= 2 * jnp.finfo(carry.right.dtype).eps)
        satisfaction = carry.log_L > log_L_constraint
        # Allow if on plateau to fly around the plateau for a while
        lesser_satisfaction = jnp.bitwise_and(seed_point.log_L0 == log_L_constraint, carry.log_L == log_L_constraint)
        done = jnp.bitwise_or(jnp.bitwise_or(close_to_zero_interval, satisfaction), lesser_satisfaction)
        return jnp.bitwise_not(done)

    def body(carry: Carry) -> Carry:
        key, t_key, shrink_key = random.split(carry.key, 3)
        left, right = _shrink_region(
            point_U=carry.point_U,
            point_U0=seed_point.U0,
            left=carry.left,
            right=carry.right
        )
        point_U = _new_sample(
            key=t_key,
            left=left,
            right=right
        )
        log_L = model.forward(point_U)
        num_likelihood_evaluations = carry.num_likelihood_evaluations + jnp.ones_like(carry.num_likelihood_evaluations)
        return Carry(
            key=key,
            left=left,
            right=right,
            point_U=point_U,
            log_L=log_L,
            num_likelihood_evaluations=num_likelihood_evaluations,
        )

    key, slice_key, t_key = random.split(key, 3)
    num_likelihood_evaluations = jnp.full((), 0, int_type)
    (left, right) = _slice_bounds(
        key=slice_key,
        point_U0=seed_point.U0,
        num_restrict_dims=num_restrict_dims
    )
    point_U = _new_sample(
        key=t_key,
        left=left,
        right=right
    )
    log_L = model.forward(point_U)
    init_carry = Carry(
        key=key,
        left=left,
        right=right,
        point_U=point_U,
        log_L=log_L,
        num_likelihood_evaluations=num_likelihood_evaluations
    )

    carry = lax.while_loop(
        cond_fun=cond,
        body_fun=body,
        init_val=init_carry
    )
    return carry.point_U, carry.log_L, carry.num_likelihood_evaluations


class MultiDimSliceSampler(BaseAbstractMarkovSampler):

    def __init__(self, model: BaseAbstractModel, num_slices: int, num_phantom_save: int,
                 num_restrict_dims: Optional[int] = None):
        """
        Multi-dimensional slice sampler, with exponential shrinkage. Produces correlated (non-i.i.d.) samples.

        Notes: Not very efficient.

        Args:
            model: AbstractModel
            num_slices: number of slices between acceptance, in units of 1, unlike other software which does it in
                units of prior dimension.
            num_phantom_save: number of phantom samples to save. Phantom samples are samples that meeting the constraint
                but are not accepted. They can be used for numerous things, e.g. to estimate the evidence uncertainty.
            num_restrict_dims: size of subspace to slice along. Setting to 1 would be like UniDimSliceSampler,
                but far less efficient.
        """
        super().__init__(model=model)
        if num_slices < 1:
            raise ValueError(f"num_slices must be > 0.")
        if num_phantom_save < 0:
            raise ValueError(f"num_phantom_save should be >= 0, got {num_phantom_save}.")
        if num_phantom_save >= num_slices:
            raise ValueError(f"num_phantom_save should be < num_slices, got {num_phantom_save} >= {num_slices}.")
        self.num_slices = int(num_slices)
        self.num_phantom_save = int(num_phantom_save)
        if num_restrict_dims is not None:
            if num_restrict_dims == 1:
                raise ValueError(f"If restricting to 1 dimension, then you should use UniDimSliceSampler.")
            if not (1 < num_restrict_dims <= model.U_ndims):
                raise ValueError(f"Expected num_restriction dim in (1, {model.U_ndims}], got {num_restrict_dims}.")
        self.num_restrict_dims = int(num_restrict_dims)

    def num_phantom(self) -> int:
        return self.num_phantom_save

    def pre_process(self, state: StaticStandardNestedSamplerState) -> SamplerState:
        sample_collection = tree_map(lambda x: x[state.front_idx], state.sample_collection)
        return (sample_collection,)

    def post_process(self, sample_collection: StaticStandardSampleCollection,
                     sampler_state: SamplerState) -> SamplerState:
        return (sample_collection,)

    def get_seed_point(self, key: PRNGKey, sampler_state: SamplerState,
                       log_L_constraint: FloatArray) -> SeedPoint:

        sample_collection: StaticStandardSampleCollection
        (sample_collection,) = sampler_state

        select_mask = sample_collection.log_L > log_L_constraint
        # If non satisfied samples, then choose randomly from them.
        any_satisfied = jnp.any(select_mask)
        yes_ = jnp.asarray(0., float_type)
        no_ = jnp.asarray(-jnp.inf, float_type)
        unnorm_select_log_prob = jnp.where(
            any_satisfied,
            jnp.where(select_mask, yes_, no_),
            yes_
        )
        # Choose randomly where mask is True
        g = random.gumbel(key, shape=unnorm_select_log_prob.shape)
        sample_idx = jnp.argmax(g + unnorm_select_log_prob)

        return SeedPoint(
            U0=sample_collection.U_samples[sample_idx],
            log_L0=sample_collection.log_L[sample_idx]
        )

    def get_sample_from_seed(self, key: PRNGKey, seed_point: SeedPoint, log_L_constraint: FloatArray,
                             sampler_state: SamplerState) -> Tuple[Sample, Sample]:

        def propose_op(sample: Sample, key: PRNGKey) -> Sample:
            U_sample, log_L, num_likelihood_evaluations = _new_proposal(
                key=key,
                seed_point=SeedPoint(
                    U0=sample.U_sample,
                    log_L0=sample.log_L
                ),
                num_restrict_dims=self.num_restrict_dims,
                log_L_constraint=log_L_constraint,
                model=self.model
            )
            return Sample(
                U_sample=U_sample,
                log_L_constraint=log_L_constraint,
                log_L=log_L,
                num_likelihood_evaluations=num_likelihood_evaluations + sample.num_likelihood_evaluations
            )

        init_sample = Sample(
            U_sample=seed_point.U0,
            log_L_constraint=log_L_constraint,
            log_L=seed_point.log_L0,
            num_likelihood_evaluations=jnp.asarray(0, int_type)
        )
        final_sample, cumulative_samples = cumulative_op_static(
            op=propose_op,
            init=init_sample,
            xs=random.split(key, self.num_slices),
            unroll=2
        )

        # Last sample is the final sample, the rest are potential phantom samples
        # Take only the last num_phantom_save phantom samples
        phantom_samples: Sample = tree_map(lambda x: x[-(self.num_phantom_save + 1):-1], cumulative_samples)

        return final_sample, phantom_samples



import logging
from functools import partial
from typing import Tuple, Dict, Any, Optional, NamedTuple

import jax
import numpy as np
import tensorflow_probability.substrates.jax as tfp
from jax import numpy as jnp, random
from jax._src.scipy.special import logsumexp
from jaxopt import NonlinearCG, ArmijoSGD
from tqdm import tqdm

from jaxns.internals.cumulative_ops import cumulative_op_static
from jaxns.internals.log_semiring import LogSpace

try:
    import haiku as hk
except ImportError:
    print("You must `pip install dm-haiku` first.")
    raise

try:
    import optax
except ImportError:
    print("You must `pip install optax` first.")
    raise

from jaxns import DefaultNestedSampler, Model
from jaxns.internals.types import TerminationCondition, NestedSamplerResults, StaticStandardNestedSamplerState, \
    IntArray, PRNGKey, float_type

__all__ = [
    'EvidenceMaximisation'
]

tfpd = tfp.distributions
tfpk = tfp.math.psd_kernels

logger = logging.getLogger('jaxns')


class MStepData(NamedTuple):
    U_samples: jnp.ndarray
    log_weights: jnp.ndarray
    # log_dp_mean: jnp.ndarray
    # log_L_samples: jnp.ndarray
    # log_Z_mean: jnp.ndarray


def next_power_2(x: int) -> int:
    """
    Next largest power of 2.

    Args:
        x:  int

    Returns:
        next largest n**2
    """
    return int(2 ** np.ceil(np.log2(x)))


class EvidenceMaximisation:
    """
    Evidence Maximisation class, that implements the E and M steps. Iteratively computes the evidence and maximises it.
    """

    def __init__(self, model: Model, ns_kwargs: Dict[str, Any],
                 max_num_epochs: int = 50, gtol=1e-2, momentum=0.9,
                 log_Z_ftol=1., log_Z_atol=1e-4,
                 batch_size: int = 128,
                 termination_cond: Optional[TerminationCondition] = None,
                 verbose: bool = False):

        """
        Initialise the EM class.

        Args:
            model: The model to train.
            max_num_epochs: The maximum number of epochs to run M-step for.
            gtol: The parameter tolerance for the M-step. End when all parameters change by less than gtol.
            log_Z_ftol, log_Z_atol: The tolerances for the change in the evidence as function of log_Z_uncert.
                Terminate if the change in log_Z is less than max(log_Z_ftol * log_Z_uncert, log_Z_atol).
            ns_kwargs: The keyword arguments to pass to the nested sampler. Needs at least `max_samples`.
            verbose: Whether to print progress verbosely.
        """
        if not isinstance(model, Model):
            raise ValueError("model must be an instance of ParametrisedModel")
        self.model = model
        self.max_num_epochs = max_num_epochs
        self.gtol = gtol
        self.momentum = momentum
        self.log_Z_ftol = log_Z_ftol
        self.log_Z_atol = log_Z_atol
        self._verbose = bool(verbose)
        self._ns_kwargs = ns_kwargs
        self._batch_size = batch_size
        self._termination_cond = termination_cond or TerminationCondition()
        self._e_step = self._create_e_step()
        self._m_step = self._create_m_step_stochastic()

    def _create_e_step(self):
        """
        Create a compiled function that runs nested sampling and returns trimmed results.

        Returns:
            A compiled function that runs nested sampling and returns trimmed results.
        """

        def _ns_solve(params: hk.MutableParams, key: random.PRNGKey) -> Tuple[
            IntArray, StaticStandardNestedSamplerState]:
            model = self.model(params=params)
            ns = DefaultNestedSampler(model=model, **self._ns_kwargs)
            termination_reason, state = ns(key)
            return termination_reason, state

        # Ahead of time compile the function
        ns_compiled = jax.jit(_ns_solve).lower(self.model.params, random.PRNGKey(42)).compile()

        def _e_step(key: PRNGKey, params: hk.MutableParams, p_bar: tqdm) -> NestedSamplerResults:
            p_bar.set_description(f"Running E-step... {p_bar.desc}")
            termination_reason, state = ns_compiled(params, key)
            ns = DefaultNestedSampler(model=self.model(params=params), **self._ns_kwargs)
            # Trim now
            return ns.to_results(termination_reason=termination_reason, state=state, trim=True)

        return _e_step

    def e_step(self, key: PRNGKey, params: hk.MutableParams, p_bar: tqdm) -> NestedSamplerResults:
        """
        The E-step is just nested sampling.

        Args:
            key: The random number generator key.
            params: The parameters to use.
            p_bar: progress bar

        Returns:
            The nested sampling results.
        """

        # The E-step is just nested sampling
        return self._e_step(key, params, p_bar)

    def _m_step_iterator(self, key: PRNGKey, data: MStepData):
        num_samples = int(data.U_samples.shape[0])
        permutation = jax.random.permutation(key, num_samples)
        num_batches = num_samples // self._batch_size
        if num_batches == 0:
            raise RuntimeError("Batch size is too large for number of samples.")
        for i in range(num_batches):
            perm = permutation[i * self._batch_size:(i + 1) * self._batch_size]
            batch = MStepData(
                U_samples=data.U_samples[perm],
                log_weights=data.log_weights[perm]
            )
            yield batch

    def _create_m_step_stochastic(self):
        def log_evidence(params: hk.MutableParams, data: MStepData):
            # Compute the log evidence
            model = self.model(params=params)
            # To make manageable, we could do chunked_pmap
            log_dZ = jax.vmap(
                lambda U, log_weight: model.forward(U) + log_weight
            )(data.U_samples, data.log_weights)
            # We add the log_Z_mean because log_dp_mean is normalised
            log_Z = logsumexp(log_dZ)
            return log_Z

        def loss(params: hk.MutableParams, data: MStepData):
            log_Z, grad = jax.value_and_grad(log_evidence, argnums=0)(params, data)
            obj = -log_Z
            grad = jax.tree_map(jnp.negative, grad)
            aux = (log_Z,)
            if self._verbose:
                jax.debug.print("log_Z={log_Z}", log_Z=log_Z)
            return (obj, aux), grad

        solver = ArmijoSGD(
            fun=loss,
            has_aux=True,
            value_and_grad=True,
            jit=True,
            unroll=False,
            verbose=self._verbose,
            momentum=self.momentum
        )

        def _m_step_stochastic(key: PRNGKey, params: hk.MutableParams, data: MStepData) -> Tuple[hk.MutableParams, Any]:
            """
            The M-step is just evidence maximisation.

            Args:
                key: The random number generator key.
                params: The parameters to use.
                data: The data to use.

            Returns:
                The updated parameters.
            """

            # The M-step is just evidence maximisation

            iterator = self._m_step_iterator(key, data)
            opt_results = solver.run_iterator(init_params=params, iterator=iterator)
            return opt_results.params, opt_results.state.aux

        return _m_step_stochastic

    def _create_m_step(self):

        def log_evidence(params: hk.MutableParams, data: MStepData):
            # Compute the log evidence
            model = self.model(params=params)

            def op(log_Z, data):
                log_dZ = model.forward(data.U_samples) + data.log_weights
                return (LogSpace(log_Z) + LogSpace(log_dZ)).log_abs_val

            log_Z, _ = cumulative_op_static(op=op, init=jnp.asarray(-jnp.inf, float_type), xs=data)
            return log_Z

        def loss(params: hk.MutableParams, data: MStepData):
            log_Z, grad = jax.value_and_grad(log_evidence, argnums=0)(params, data)
            obj = -log_Z
            grad = jax.tree_map(jnp.negative, grad)
            aux = (log_Z,)
            if self._verbose:
                jax.debug.print("log_Z={log_Z}", log_Z=log_Z)
            return (obj, aux), grad

        solver = NonlinearCG(
            fun=loss,
            has_aux=True,
            value_and_grad=True,
            jit=True,
            unroll=False,
            verbose=self._verbose
        )

        @partial(jax.jit, static_argnums=(0,))
        def _m_step(key: PRNGKey, params: hk.MutableParams, data: MStepData) -> Tuple[hk.MutableParams, Any]:
            """
            The M-step is just evidence maximisation.

            Args:
                params: The parameters to use.
                data: The data to use.

            Returns:
                The updated parameters and the negative log evidence.
            """
            opt_results = solver.run(init_params=params, data=data)
            return opt_results.params, opt_results.state.aux

        return _m_step

    def m_step(self, key: PRNGKey, params: hk.MutableParams, ns_results: NestedSamplerResults, p_bar: tqdm) -> Tuple[
        hk.MutableParams, Any]:
        """
        The M-step is just evidence maximisation. We pad the data to the next power of 2, to make JIT compilation
        happen less frequently.

        Args:
            key: The random number generator key.
            params: The parameters to use.
            ns_results: The nested sampling results to use.
            p_bar: progress bar

        Returns:
            The updated parameters
        """
        # next_power_2 pad
        num_samples = int(ns_results.total_num_samples)
        n = next_power_2(num_samples)

        p_bar.set_description(f"Running M-step ({num_samples} samples padded to {n})... {p_bar.desc}")

        def _pad_to_n(x, fill_value, dtype):
            return jnp.concatenate([x, jnp.full((n - x.shape[0],) + x.shape[1:], fill_value, dtype)], axis=0)

        log_weights = ns_results.log_dp_mean - ns_results.log_L_samples + ns_results.log_Z_mean
        data = MStepData(
            U_samples=_pad_to_n(ns_results.U_samples, 0.5, float_type),
            log_weights=_pad_to_n(log_weights, -jnp.inf, float_type)
        )
        desc = p_bar.desc
        last_params = params
        epoch = 0
        log_Z = None
        while epoch < self.max_num_epochs:
            params, (log_Z,) = self._m_step(key=key, params=params, data=data)
            l_oo = jax.tree_map(lambda x, y: jnp.max(jnp.abs(x - y)), last_params, params)
            last_params = params
            p_bar.set_description(f"{desc}: Epoch {epoch}: log_Z={log_Z}, l_oo={l_oo}")
            if all(_l_oo < self.gtol for _l_oo in jax.tree_leaves(l_oo)):
                break
            epoch += 1

        return params, log_Z

    def train(self, num_steps: int = 10, params: Optional[hk.MutableParams] = None) -> \
            Tuple[
                NestedSamplerResults, hk.MutableParams]:
        """
        Train the model using EM for num_steps.

        Args:
            num_steps: The number of steps to train for, or until convergence.
            params: The initial parameters to use. If None, then the model's params are used.
            do_final_e_step: Whether to do a final E-step after training, for updated evidence.

        Returns:
            The trained parameters.
        """
        if params is None:
            params = self.model.params

        log_Z = -jnp.inf

        # Initialize the progress bar with a description
        p_bar = tqdm(range(num_steps), desc="Processing Steps", dynamic_ncols=True)

        ns_results = None
        for step in p_bar:
            key_e_stek, key_m_step = random.split(random.PRNGKey(step), 2)

            # Execute the e_step
            if ns_results is None:
                p_bar.set_description(f"Step {step}: Initial run")
            else:
                p_bar.set_description(
                    f"Step {step}: log Z = {ns_results.log_Z_mean:.4f} +- {ns_results.log_Z_uncert:.4f}"
                )
            ns_results = self.e_step(key=key_e_stek, params=params, p_bar=p_bar)
            # Update progress bar description

            # Check termination condition
            log_Z_change = jnp.abs(ns_results.log_Z_mean - log_Z)
            if log_Z_change < max(self.log_Z_ftol * ns_results.log_Z_uncert, self.log_Z_atol):
                p_bar.set_description(f"Convergence achieved at step {step}.")
                break

            # Update log_Z and log_Z_uncert values
            log_Z = ns_results.log_Z_mean

            # Execute the m_step
            p_bar.set_description(
                f"Step {step}: log Z = {ns_results.log_Z_mean:.4f} +- {ns_results.log_Z_uncert:.4f}"
            )
            params, log_Z_opt = self.m_step(key=key_m_step, params=params, ns_results=ns_results, p_bar=p_bar)

        if ns_results is None:
            raise RuntimeError("No results were computed.")
        return ns_results, params



import io
import logging
from typing import NamedTuple, TextIO, Union, Optional, Tuple, TypeVar, Callable

import numpy as np
from jax import numpy as jnp, tree_map, vmap, random, jit, lax

from jaxns.framework.bases import BaseAbstractModel
from jaxns.internals.cumulative_ops import cumulative_op_static
from jaxns.internals.log_semiring import LogSpace
from jaxns.internals.maps import prepare_func_args
from jaxns.internals.random import resample_indicies
from jaxns.internals.types import NestedSamplerResults, float_type, XType, UType, FloatArray, IntArray
from jaxns.internals.types import PRNGKey
from jaxns.warnings import deprecated

logger = logging.getLogger('jaxns')

__all__ = [
    'resample',
    'marginalise_static_from_U',
    'marginalise_dynamic_from_U',
    'marginalise_static',
    'marginalise_dynamic',
    'maximum_a_posteriori_point',
    'evaluate_map_estimate',
    'summary',
    'analytic_posterior_samples',
    'sample_evidence',
    'bruteforce_posterior_samples',
    'bruteforce_evidence',
    'save_pytree',
    'save_results',
    'load_pytree',
    'load_results'
]


def resample(key: PRNGKey, samples: Union[XType, UType], log_weights: jnp.ndarray, S: int = None,
             replace: bool = False) -> XType:
    """
    Resample the weighted samples into uniformly weighted samples.

    Args:
        key: PRNGKey
        samples: samples from nested sampled results
        log_weights: log-posterior weight
        S: number of samples to generate. Will use Kish's estimate of ESS if None.
        replace: whether to sample with replacement

    Returns:
        equally weighted samples
    """
    idx = resample_indicies(key, log_weights, S=S, replace=replace)
    return tree_map(lambda s: s[idx, ...], samples)


_V = TypeVar('_V')


def evaluate_map_estimate_from_U(results: NestedSamplerResults, model: BaseAbstractModel, fun: Callable[..., _V]) -> _V:
    """
    Marginalises function over posterior samples, where ESS is static.

    Args:
        results: results from run
        fun (:code:`callable(**kwargs)`): function to marginalise

    Returns:
        estimate at MAP sample point
    """
    map_sample_U = maximum_a_posteriori_point_U(results=results)
    V = model.prepare_input(U=map_sample_U)
    return fun(*V)


def marginalise_static_from_U(key: PRNGKey, U_samples: UType, model: BaseAbstractModel, log_weights: jnp.ndarray,
                              ESS: int,
                              fun: Callable[..., _V]) -> _V:
    """
    Marginalises function over posterior samples, where ESS is static.

    Args:
        key: PRNG key
        U_samples: array of U samples
        model: model
        log_weights: log weights from nested sampling
        ESS: static effective sample size
        fun (:code:`callable(**kwargs)`): function to marginalise

    Returns:
        expectation over resampled samples
    """
    U_samples = resample(key, U_samples, log_weights, S=ESS, replace=True)

    def _eval(U):
        V = model.prepare_input(U=U)
        return fun(*V)

    marginalised = tree_map(lambda marg: jnp.nanmean(marg, axis=0), vmap(_eval)(U_samples))
    return marginalised


def marginalise_dynamic_from_U(key: PRNGKey, U_samples: UType, model: BaseAbstractModel, log_weights: jnp.ndarray,
                               ESS: jnp.ndarray,
                               fun: Callable[..., _V]) -> _V:
    """
    Marginalises function over posterior samples, where ESS can be dynamic.

    Args:
        key: PRNG key
        U_samples: array of U samples
        model: model
        log_weights: log weights from nested sampling
        ESS: dynamic effective sample size
        fun (:code:`callable(**kwargs)`): function to marginalise

    Returns:
        expectation of `func` over resampled samples.
    """
    ESS = jnp.asarray(ESS)

    def _eval(U):
        V = model.prepare_input(U=U)
        return fun(*V)

    def body(state):
        (key, i, count, marginalised) = state
        key, resample_key = random.split(key, 2)
        _samples = resample(resample_key, U_samples, log_weights, S=1)
        _sample = tree_map(lambda v: v[0], _samples)
        update = _eval(_sample)
        count = tree_map(lambda y, c: jnp.where(jnp.any(jnp.isnan(y)), c, c + jnp.asarray(1, c.dtype)),
                         update, count)
        marginalised = tree_map(lambda x, y: jnp.where(jnp.isnan(y), x, x + y.astype(x.dtype)),
                                marginalised, update)
        return (key, i + jnp.ones_like(i), count, marginalised)

    test_output = fun(**tree_map(lambda v: v[0], U_samples))
    count = tree_map(lambda x: jnp.asarray(0, x.dtype), test_output)
    init_marginalised = tree_map(lambda x: jnp.zeros_like(x), test_output)
    (_, _, count, marginalised) = lax.while_loop(lambda state: state[1] < ESS,
                                                 body,
                                                 (key, jnp.array(0, ESS.dtype), count, init_marginalised))
    marginalised = tree_map(lambda x, c: x / c, marginalised, count)
    return marginalised


def marginalise_static(key: PRNGKey, samples: XType, log_weights: jnp.ndarray, ESS: int, fun: Callable[..., _V]) -> _V:
    """
    Marginalises function over posterior samples, where ESS is static.

    Args:
        key: PRNG key
        samples (dict): dict of batched array of nested sampling samples
        log_weights: log weights from nested sampling
        ESS: static effective sample size
        fun (:code:`callable(**kwargs)`): function to marginalise

    Returns:
        expectation over resampled samples
    """
    fun = prepare_func_args(fun)
    samples = resample(key, samples, log_weights, S=ESS, replace=True)
    marginalised = tree_map(lambda marg: jnp.nanmean(marg, axis=0), vmap(fun)(**samples))
    return marginalised


def marginalise_dynamic(key: PRNGKey, samples: XType, log_weights: jnp.ndarray, ESS: jnp.ndarray,
                        fun: Callable[..., _V]) -> _V:
    """
    Marginalises function over posterior samples, where ESS can be dynamic.

    Args:
        key: PRNG key
        samples (dict): dict of batched array of nested sampling samples
        log_weights: log weights from nested sampling
        ESS: dynamic effective sample size
        fun (:code:`callable(**kwargs)`): function to marginalise

    Returns:
        expectation of `func` over resampled samples.
    """
    fun = prepare_func_args(fun)
    ESS = jnp.asarray(ESS)

    def body(state):
        (key, i, count, marginalised) = state
        key, resample_key = random.split(key, 2)
        _samples = resample(resample_key, samples, log_weights, S=1)
        _sample = tree_map(lambda v: v[0], _samples)
        update = fun(**_sample)
        count = tree_map(lambda y, c: jnp.where(jnp.any(jnp.isnan(y)), c, c + jnp.asarray(1, c.dtype)),
                         update, count)
        marginalised = tree_map(lambda x, y: jnp.where(jnp.isnan(y), x, x + y.astype(x.dtype)),
                                marginalised, update)
        return (key, i + jnp.ones_like(i), count, marginalised)

    test_output = fun(**tree_map(lambda v: v[0], samples))
    count = tree_map(lambda x: jnp.asarray(0, x.dtype), test_output)
    init_marginalised = tree_map(lambda x: jnp.zeros_like(x), test_output)
    (_, _, count, marginalised) = lax.while_loop(lambda state: state[1] < ESS,
                                                 body,
                                                 (key, jnp.array(0, ESS.dtype), count, init_marginalised))
    marginalised = tree_map(lambda x, c: x / c, marginalised, count)
    return marginalised


def maximum_a_posteriori_point(results: NestedSamplerResults) -> XType:
    """
    Get the MAP point of a nested sampling result.
    Does this by choosing the point with largest L(x) p(x).

    Args:
        results (NestedSamplerResult): Nested sampler result

    Returns:
        dict of samples at MAP-point.
    """

    map_idx = jnp.argmax(results.log_posterior_density)
    map_points = tree_map(lambda x: x[map_idx], results.samples)
    return map_points


def maximum_a_posteriori_point_U(results: NestedSamplerResults) -> UType:
    """
    Get the MAP point of a nested sampling result.
    Does this by choosing the point with largest L(x) p(x).

    Args:
        results (NestedSamplerResult): Nested sampler result

    Returns:
        dict of samples at MAP-point.
    """

    map_idx = jnp.argmax(results.log_posterior_density)
    map_points = tree_map(lambda x: x[map_idx], results.U_samples)
    return map_points


def evaluate_map_estimate(results: NestedSamplerResults, fun: Callable[..., _V]) -> _V:
    """
    Marginalises function over posterior samples, where ESS is static.

    Args:
        results: results from run
        fun (:code:`callable(**kwargs)`): function to marginalise

    Returns:
        estimate at MAP sample point
    """
    fun = prepare_func_args(fun)
    map_sample = maximum_a_posteriori_point(results=results)
    return fun(**map_sample)


def _bit_mask(int_mask, width=8):
    """
    Convert an integer mask into a bit-mask. I.e. convert an integer into list of left-starting bits.

    Examples:

    1 -> [1,0,0,0,0,0,0,0]
    2 -> [0,1,0,0,0,0,0,0]
    3 -> [1,1,0,0,0,0,0,0]

    Args:
        int_mask: int
        width: number of output bits

    Returns:
        List of bits from left
    """
    return list(map(int, '{:0{size}b}'.format(int_mask, size=width)))[::-1]


def summary(results: NestedSamplerResults, with_parametrised: bool = False, f_obj: Optional[Union[str, TextIO]] = None):
    """
    Gives a summary of the results of a nested sampling run.

    Args:
        results (NestedSamplerResults): Nested sampler result
        with_parametrised: whether to include parametrised samples
        f_obj: file-like object to write summary to. If None, prints to stdout.
    """
    main_s = []

    def _print(s):
        print(s)
        main_s.append(s)

    def _round(v, uncert_v):
        v = float(v)
        uncert_v = float(uncert_v)
        try:
            sig_figs = -int("{:e}".format(uncert_v).split('e')[1]) + 1
            return round(float(v), sig_figs)
        except:
            return float(v)

    def _print_termination_reason(_termination_reason: int):
        termination_bit_mask = _bit_mask(int(_termination_reason), width=8)

        for bit, condition in zip(termination_bit_mask, [
            'Reached max samples',
            'Evidence uncertainty low enough',
            'Small remaining evidence',
            'Reached ESS',
            "Used max num likelihood evaluations",
            'Likelihood contour reached',
            'Sampler efficiency too low',
            'All live-points are on a single plateau (potential numerical errors, consider 64-bit)'
        ]):
            if bit == 1:
                _print(condition)

    _print("--------")
    _print("Termination Conditions:")
    if np.size(results.termination_reason) > 1:  # Reasons for each parallel sampler
        print(results.termination_reason)
        for sampler_idx in range(np.size(results.termination_reason)):
            _print(f"Sampler {sampler_idx}:")
            _print_termination_reason(int(results.termination_reason[sampler_idx]))
    else:
        _print_termination_reason(int(results.termination_reason))
    _print("--------")
    _print(f"likelihood evals: {int(results.total_num_likelihood_evaluations):d}")
    _print(f"samples: {int(results.total_num_samples):d}")
    _print(f"phantom samples: {int(results.total_phantom_samples):d}")
    _print(
        f"likelihood evals / sample: {float(results.total_num_likelihood_evaluations / results.total_num_samples):.1f}"
    )
    _print(
        f"phantom fraction (%): {100 * float(results.total_phantom_samples / results.total_num_samples):.1f}%"
    )
    _print("--------")
    _print(
        f"logZ={_round(results.log_Z_mean, results.log_Z_uncert)} +- {_round(results.log_Z_uncert, results.log_Z_uncert)}"
    )
    # _print("H={} +- {}".format(
    #     _round(results.H_mean, results.H_uncert), _round(results.H_uncert, results.H_uncert)))
    _print(
        f"H={_round(results.H_mean, 0.1)}"
    )
    _print(
        f"ESS={int(results.ESS):d}"
    )

    samples = results.samples
    if with_parametrised:
        samples.update(results.parametrised_samples)

    max_like_idx = np.argmax(results.log_L_samples)
    max_like_points = tree_map(lambda x: x[max_like_idx], samples)

    uniform_samples = resample(random.PRNGKey(23426),
                               samples,
                               results.log_dp_mean,
                               S=max(100, int(results.ESS)),
                               replace=True)

    max_map_idx = np.argmax(results.log_posterior_density)
    map_points = tree_map(lambda x: x[max_map_idx], results.samples)

    for name in uniform_samples.keys():
        _samples = uniform_samples[name].reshape((uniform_samples[name].shape[0], -1))
        _max_like_points = max_like_points[name].reshape((-1,))
        _map_points = map_points[name].reshape((-1,))
        ndims = _samples.shape[1]
        _print("--------")
        var_name = name if ndims == 1 else "{}[#]".format(name)
        _print(
            f"{var_name}: mean +- std.dev. | 10%ile / 50%ile / 90%ile | MAP est. | max(L) est."
        )
        for dim in range(ndims):
            _uncert = np.std(_samples[:, dim])
            _max_like_point = _max_like_points[dim]
            _map_point = _map_points[dim]
            # two sig-figs based on uncert
            sig_figs = -int("{:e}".format(_uncert).split('e')[1]) + 1

            def _round(ar):
                return round(float(ar), sig_figs)

            _uncert = _round(_uncert)
            _print("{}: {} +- {} | {} / {} / {} | {} | {}".format(
                name if ndims == 1 else "{}[{}]".format(name, dim),
                _round(np.mean(_samples[:, dim])), _uncert,
                *[_round(a) for a in np.percentile(_samples[:, dim], np.asarray([10, 50, 90]))],
                _round(_map_point),
                _round(_max_like_point)
            ))
    _print("--------")
    if f_obj is not None:
        out = "\n".join(main_s)
        if isinstance(f_obj, str):
            with open(f_obj, 'w') as f:
                f.write(out)
        elif isinstance(f_obj, io.TextIOBase):
            f_obj.write(out)
        else:
            raise TypeError(f"Invalid f_obj: {type(f_obj)}")


def sample_evidence(key: PRNGKey, num_live_points_per_sample: IntArray, log_L_samples: FloatArray,
                    S: int = 100) -> FloatArray:
    """
    Sample the evidence distribution, but stochastically simulating the shrinkage distribution.

    Note: this produces approximate samples, since there is also an uncertainty in the placement of the contours during
    shrinkage. Incorporating this stochasticity into the simulation would require running an entire
    nested sampling many times.

    Args:
        key: PRNGKey
        num_live_points_per_sample: the number of live points for each sample
        log_L_samples: the log-L of samples
        S: The number of samples to produce

    Returns:
        samples of log(Z)
    """

    def accumulate_op(accumulate, y):
        # T ~ Beta(n[i],1) <==> T ~ Kumaraswamy(n[i],1)
        (key, num_live_points, log_L) = y
        (log_Z, log_X) = accumulate
        log_T = jnp.log(random.uniform(key, num_live_points.shape, dtype=log_L.dtype)) / num_live_points
        # log_T = jnp.where(num_live_points == 0., -jnp.inf, log_T)
        # log_T = jnp.log(random.beta(key, num_live_points, 1.))
        T = LogSpace(log_T)
        X = LogSpace(log_X)
        Z = LogSpace(log_Z)
        L = LogSpace(log_L)
        next_X = X * T
        dZ = (X - next_X) * L
        next_Z = Z + dZ
        return (next_Z.log_abs_val, next_X.log_abs_val)

    def single_log_Z_sample(key: PRNGKey) -> FloatArray:
        init = (jnp.asarray(-jnp.inf, log_L_samples.dtype), jnp.asarray(0., log_L_samples.dtype))
        xs = (random.split(key, num_live_points_per_sample.shape[0]), num_live_points_per_sample, log_L_samples)
        final_accumulate, _ = cumulative_op_static(accumulate_op, init=init, xs=xs)
        (log_Z, _) = final_accumulate
        return log_Z

    log_Z_samples = vmap(single_log_Z_sample)(random.split(key, S))
    return log_Z_samples


def bruteforce_posterior_samples(model: BaseAbstractModel, S: int = 60) -> Tuple[XType, jnp.ndarray]:
    """
    Compute the posterior with brute-force over a regular grid.

    Args:
        model: model
        S: resolution of grid

    Returns:
        samples, and log-weight
    """
    u_vec = jnp.linspace(jnp.finfo(float_type).eps, 1. - jnp.finfo(float_type).eps, S)
    du = u_vec[1] - u_vec[0]
    args = jnp.stack([x.flatten() for x in jnp.meshgrid(*[u_vec] * model.U_ndims, indexing='ij')], axis=-1)
    samples = jit(vmap(model.transform))(args)
    log_L = jit(vmap(model.forward))(args)
    dZ = LogSpace(log_L) * LogSpace(jnp.log(du)) ** model.U_ndims
    return samples, dZ.log_abs_val


def bruteforce_evidence(model: BaseAbstractModel, S: int = 60):
    """
    Compute the evidence with brute-force over a regular grid.

    Args:
        model: model
        S: resolution of grid

    Returns:
        log(Z)
    """

    u_vec = jnp.linspace(jnp.finfo(float_type).eps, 1. - jnp.finfo(float_type).eps, S)
    du = u_vec[1] - u_vec[0]
    args = jnp.stack([x.flatten() for x in jnp.meshgrid(*[u_vec] * model.U_ndims, indexing='ij')], axis=-1)
    Z_true = (LogSpace(jit(vmap(model.forward))(args)).nansum() * LogSpace(
        jnp.log(du)) ** model.U_ndims)
    return Z_true.log_abs_val


@deprecated(bruteforce_posterior_samples)
def analytic_posterior_samples(model: BaseAbstractModel, S: int = 60):
    """
    Compute the evidence with brute-force over a regular grid.

    Args:
        model: model
        S: resolution of grid

    Returns:
        log(Z)
    """
    logger.warning(f"")

    u_vec = jnp.linspace(jnp.finfo(float_type).eps, 1. - jnp.finfo(float_type).eps, S)
    du = u_vec[1] - u_vec[0]
    args = jnp.stack([x.flatten() for x in jnp.meshgrid(*[u_vec] * model.U_ndims, indexing='ij')], axis=-1)
    samples = jit(vmap(model.transform))(args)
    log_L = jit(vmap(model.forward))(args)
    dZ = LogSpace(log_L) * LogSpace(jnp.log(du)) ** model.U_ndims
    return samples, dZ.log_abs_val


def _isinstance_namedtuple(obj) -> bool:
    return (
            isinstance(obj, tuple) and
            hasattr(obj, '_asdict') and
            hasattr(obj, '_fields')
    )


def save_pytree(pytree: NamedTuple, save_file: str):
    """
    Saves results of nested sampler in a npz file.

    Args:
        pytree: Nested sampler result
        save_file: filename
    """
    pytree_np = tree_map(lambda v: np.asarray(v) if v is not None else None, pytree)

    def _pytree_asdict(pytree):
        _data_dict = pytree._asdict()
        data_dict = {}
        for k, v in _data_dict.items():
            if _isinstance_namedtuple(v):
                data_dict[k] = _pytree_asdict(v)
            elif isinstance(v, (dict, np.ndarray, None.__class__)):
                data_dict[k] = v
            else:
                raise ValueError("key, value pair {}, {} unknown".format(k, v))
        return data_dict

    data_dict = _pytree_asdict(pytree_np)
    np.savez(save_file, **data_dict)


def save_results(results: NestedSamplerResults, save_file: str):
    """
    Saves results of nested sampler in a npz file.

    Args:
        results (NestedSamplerResults): Nested sampler result
        save_file (str): filename
    """
    save_pytree(results, save_file)


def load_pytree(save_file: str):
    """
    Loads saved nested sampler results from a npz file.

    Args:
        save_file (str): filename

    Returns:
        NestedSamplerResults
    """
    _data_dict = np.load(save_file, allow_pickle=True)
    data_dict = {}
    for k, v in _data_dict.items():
        if v.size == 1:
            if v.item() is None:
                data_dict[k] = None
            else:
                data_dict[k] = tree_map(lambda v: jnp.asarray(v), v.item())
        else:
            data_dict[k] = jnp.asarray(v)

    return data_dict


def load_results(save_file: str) -> NestedSamplerResults:
    """
    Loads saved nested sampler results from a npz file.

    Args:
        save_file (str): filename

    Returns:
        NestedSamplerResults
    """
    data_dict = load_pytree(save_file)
    return NestedSamplerResults(**data_dict)



from typing import Union, Literal

from jax import numpy as jnp, lax
from jax.scipy.special import logsumexp

from jaxns.internals.types import SignedLog, float_type


def logaddexp(x1, x2):
    """
    Equivalent to logaddexp but supporting complex arguments.

    see np.logaddexp
    """
    if is_complex(x1) or is_complex(x2):
        select1 = x1.real > x2.real
        amax = jnp.where(select1, x1, x2)
        delta = jnp.where(select1, x2 - x1, x1 - x2)
        return jnp.where(jnp.isnan(delta),
                         x1 + x2,  # NaNs or infinities of the same sign.
                         amax + jnp.log1p(jnp.exp(delta)))
    else:
        return jnp.logaddexp(x1, x2)


def signed_logaddexp(log_abs_val1, sign1, log_abs_val2, sign2):
    r"""
    Equivalent of logaddexp but for signed quantities too.
    Broadcasting supported.

    Args:
        log_abs_val1: Logarithm of absolute value of val1, :math:`\log(|x_1|)`
        sign1: Sign of val1, :math:`\mathrm{sign}(x_1)`
        log_abs_val2: Logarithm of absolute value of val2, :math:`\log(|x_2|)`
        sign2: Sign of val2, :math:`\mathrm{sign}(x_2)`

    Returns:
        (:math:`\log(|x_1+x_2|)`, :math:`\mathrm{sign}(x_1+x_2)`)
    """
    amax = jnp.maximum(log_abs_val1, log_abs_val2)
    signmax = jnp.where(log_abs_val1 > log_abs_val2, sign1, sign2)
    delta = -jnp.abs(log_abs_val2 - log_abs_val1)  # nan iff inf - inf
    sign = sign1 * sign2
    return jnp.where(jnp.isnan(delta),
                     log_abs_val1 + log_abs_val2,  # NaNs or infinities of the same sign.
                     amax + jnp.log1p(sign * jnp.exp(delta))), signmax


def cumulative_logsumexp(u, sign=None, reverse=False, axis=0):
    if sign is not None:
        u, sign = jnp.broadcast_arrays(u, sign)

    def body(state, X):
        if sign is not None:
            (u, u_sign) = X
            (accumulant, accumulant_sign) = state
            new_accumulant, new_accumulant_sign = signed_logaddexp(accumulant, accumulant_sign, u, u_sign)
            return (new_accumulant, accumulant_sign), (new_accumulant, accumulant_sign)
        else:
            u = X
            accumulant = state
            new_accumulant = jnp.logaddexp(accumulant, u)
            return new_accumulant, new_accumulant

    if sign is not None:
        if axis != 0:
            sign = jnp.swapaxes(sign, axis, 0)
            u = jnp.swapaxes(u, axis, 0)
        state = (-jnp.inf * jnp.ones(u.shape[1:], dtype=u.dtype), jnp.ones(u.shape[1:], dtype=u.dtype))
        X = (u, sign)
    else:
        if axis != 0:
            u = jnp.swapaxes(u, axis, 0)
        state = -jnp.inf * jnp.ones(u.shape[1:], dtype=u.dtype)
        X = u
    _, result = lax.scan(body,
                         state,
                         X,
                         reverse=reverse)
    if sign is not None:
        v, v_sign = result
        if axis != 0:
            v = jnp.swapaxes(v, axis, 0)
            v_sign = jnp.swapaxes(v_sign, axis, 0)
        return v, v_sign
    else:
        v = result
        if axis != 0:
            v = jnp.swapaxes(v, axis, 0)
        return v


class LogSpace(object):
    def __init__(self, log_abs_val: Union[jnp.ndarray, float], sign: Union[jnp.ndarray, float] = None):
        self._log_abs_val = jnp.asarray(log_abs_val, float_type)
        if sign is None:
            self._sign = jnp.asarray(1., float_type)
            self._naked = True
        else:
            self._sign = jnp.asarray(sign, float_type)
            self._naked = False

    @property
    def dtype(self):
        return self.log_abs_val.dtype

    @property
    def log_abs_val(self):
        return self._log_abs_val

    @property
    def sign(self):
        return self._sign

    @property
    def value(self):
        if self._naked:
            return jnp.exp(self.log_abs_val)
        return self.sign * jnp.exp(self.log_abs_val)

    def __neg__(self):
        if self._naked:
            return LogSpace(self.log_abs_val, -jnp.ones_like(self.log_abs_val))
        return LogSpace(self.log_abs_val, -self.sign)

    def __add__(self, other):
        """
        Implements addition in log space

            log(exp(log_A) + exp(log_B))

        Args:
            other: ndarray or LogSpace, if ndarray assumed to be log(B)

        Returns:
             LogSpace
        """
        if not isinstance(other, LogSpace):
            raise TypeError(f"Expected type {type(self)} got {type(other)}")
        if self._naked and other._naked:  # no coefficients
            return LogSpace(jnp.logaddexp(self._log_abs_val, other._log_abs_val))
        return LogSpace(*signed_logaddexp(self._log_abs_val, self._sign, other._log_abs_val, other._sign))

    def __sub__(self, other):
        """
        Implements addition in log space

            log(exp(log_A) - exp(log_B))

        Args:
            other: ndarray or LogSpace, if ndarray assumed to be log(B)

        Returns:
             LogSpace
        """
        if not isinstance(other, LogSpace):
            raise TypeError(f"Expected type {type(self)} got {type(other)}")
        return LogSpace(*signed_logaddexp(self._log_abs_val, self._sign, other._log_abs_val, -other._sign))

    def __mul__(self, other):
        """
        Implements addition in log space

            log(exp(log_A) * exp(log_B))

        Args:
            other: ndarray or LogSpace, if ndarray assumed to be log(B)

        Returns:
             LogSpace
        """
        if not isinstance(other, LogSpace):
            raise TypeError(f"Expected type {type(self)} got {type(other)}")
        if self._naked and other._naked:  # no coefficients
            return LogSpace(self._log_abs_val + other._log_abs_val)
        return LogSpace(self._log_abs_val + other._log_abs_val, self._sign * other._sign)

    def __repr__(self):
        if self._naked:
            return f"LogSpace({self.log_abs_val})"
        return f"LogSpace({self.log_abs_val}, {self.sign})"

    def sum(self, axis=-1, keepdims=False):
        if not self._naked:  # no coefficients
            return LogSpace(*logsumexp(self.log_abs_val, b=self.sign, axis=axis, keepdims=keepdims, return_sign=True))
        return LogSpace(logsumexp(self._log_abs_val, axis=axis, keepdims=keepdims))

    def nansum(self, axis=-1, keepdims=False):
        log_abs_val = jnp.where(jnp.isnan(self.log_abs_val), -jnp.inf, self.log_abs_val)
        if not self._naked:  # no coefficients
            return LogSpace(*logsumexp(log_abs_val, b=self.sign, axis=axis, keepdims=keepdims, return_sign=True))
        return LogSpace(logsumexp(log_abs_val, axis=axis, keepdims=keepdims))

    def cumsum(self, axis=0, reverse=False):
        if not self._naked:  # no coefficients
            return LogSpace(*cumulative_logsumexp(self.log_abs_val, sign=self.sign, axis=axis, reverse=reverse))
        return LogSpace(cumulative_logsumexp(self._log_abs_val, axis=axis, reverse=reverse))

    def cumprod(self, axis=0):
        if not self._naked:  # no coefficients
            log_abs_val, sign = jnp.broadcast_arrays(self.log_abs_val, self.sign)
            return LogSpace(jnp.cumsum(log_abs_val, axis=axis), jnp.cumprod(sign, axis=axis))
        return LogSpace(jnp.cumsum(self._log_abs_val, axis=axis))

    def mean(self, axis=-1, keepdims=False):
        N = self._log_abs_val.shape[axis]
        return self.sum(axis=axis, keepdims=keepdims) / LogSpace(jnp.log(N))

    def var(self, axis=-1, keepdims=False):
        return (self - self.mean(axis=axis, keepdims=True)).mean(axis=axis, keepdims=keepdims)

    def log(self):
        assert self._naked
        return LogSpace(jnp.log(jnp.abs(self.log_abs_val)), jnp.sign(self.log_abs_val))

    def exp(self):
        return LogSpace(self.value)

    def sqrt(self):
        return self ** 0.5

    def abs(self):
        return LogSpace(self.log_abs_val)

    def diff(self):
        if self._naked:
            log_abs_val, sign = jnp.broadcast_arrays(self.log_abs_val, self.sign)
            return LogSpace(log_abs_val[1:], sign[1:]) - LogSpace(log_abs_val[:-1], sign[:-1])
        else:
            return LogSpace(self.log_abs_val[1:]) - LogSpace(self.log_abs_val[:-1])

    def square(self):
        return self * self

    def argmax(self):
        return jnp.argmax(self.log_abs_val)

    def maximum(self, other: "LogSpace"):
        assert self._naked and other._naked
        return LogSpace(jnp.maximum(self.log_abs_val, other.log_abs_val))

    def minimum(self, other: "LogSpace"):
        assert self._naked and other._naked
        return LogSpace(jnp.minimum(self.log_abs_val, other.log_abs_val))

    def max(self):
        assert self._naked
        return LogSpace(jnp.max(self.log_abs_val))

    def min(self):
        assert self._naked
        return LogSpace(jnp.min(self.log_abs_val))

    def concatenate(self, other: "LogSpace", axis=0):
        if self._naked and other._naked:
            return LogSpace(jnp.concatenate([self.log_abs_val, other.log_abs_val], axis=axis))
        log_abs_val, sign = jnp.broadcast_arrays(self.log_abs_val, self.sign)
        _log_abs_val, _sign = jnp.broadcast_arrays(other.log_abs_val, other.sign)
        return LogSpace(jnp.concatenate([log_abs_val, _log_abs_val], axis=axis),
                        jnp.concatenate([sign, _sign], axis=axis))

    def __getitem__(self, item):
        if self._naked:
            return LogSpace(self.log_abs_val[item])
        log_abs_val, sign = jnp.broadcast_arrays(self.log_abs_val, self.sign)
        return LogSpace(log_abs_val[item], sign[item])

    @property
    def signed_log(self):
        return SignedLog(self.log_abs_val, self.sign)

    def __gt__(self, other):
        if not isinstance(other, LogSpace):
            raise TypeError(f"Expected type {type(self)} got {type(other)}")
        if self._naked and other._naked:
            return self.log_abs_val > other.log_abs_val
        return (self / other).value > 1.

    def __lt__(self, other):
        if not isinstance(other, LogSpace):
            raise TypeError(f"Expected type {type(self)} got {type(other)}")
        if self._naked and other._naked:
            return self.log_abs_val < other.log_abs_val
        return (self / other).value < 1.

    def __ge__(self, other):
        if not isinstance(other, LogSpace):
            raise TypeError(f"Expected type {type(self)} got {type(other)}")
        if self._naked and other._naked:
            return self.log_abs_val >= other.log_abs_val
        return (self / other).value >= 1.

    def __le__(self, other):
        if not isinstance(other, LogSpace):
            raise TypeError(f"Expected type {type(self)} got {type(other)}")
        if self._naked and other._naked:
            return self.log_abs_val <= other.log_abs_val
        return (self / other).value <= 1.

    @property
    def size(self):
        if self._naked:
            return self.log_abs_val.size
        log_abs_val, sign = jnp.broadcast_arrays(self.log_abs_val, self.sign)
        return log_abs_val.size

    def __pow__(self, n):
        """
        Implements power in log space

            log(exp(log_A)**n)

        Args:
            n: int or float

        Returns:
             LogSpace
        """
        if not isinstance(n, (int, float, jnp.ndarray)):
            raise NotImplementedError("Not implemented for non-int powers.")
        n = jnp.asarray(n, float_type)
        if self._naked:
            return LogSpace(n * self.log_abs_val)
        # complex values can occur if n is not even
        return LogSpace(n * self.log_abs_val, sign=self.sign ** n)
        # for _ in range(n-1):
        #     output = output * self
        # return output

    def __truediv__(self, other):
        """
        Implements addition in log space

            log(exp(log_A) / exp(log_B))

        Args:
            other: ndarray or LogSpace, if ndarray assumed to be log(B)

        Returns:
             LogSpace
        """
        if not isinstance(other, LogSpace):
            raise TypeError(f"Expected type {type(self)} got {type(other)}")
        if self._naked and other._naked:  # no coefficients
            return LogSpace(self._log_abs_val - other._log_abs_val)
        return LogSpace(self._log_abs_val - other._log_abs_val, self._sign * other._sign)


def is_complex(a):
    return a.dtype in [jnp.complex64, jnp.complex128]


def normalise_log_space(x: LogSpace, norm_type: Literal['sum', 'max'] = 'sum') -> LogSpace:
    """
    Safely normalise a LogSpace, accounting for zero-sum.

    Args:
        x: LogSpace to normalise
        norm_type: 'sum' or 'max' normalisation

    Returns:
        normalised LogSpace
    """
    if norm_type == 'sum':
        norm = x.sum()
    elif norm_type == 'max':
        norm = x.max()
    else:
        raise ValueError(f"Unknown norm_type {norm_type}")
    x /= norm
    x = LogSpace(jnp.where(jnp.isneginf(norm.log_abs_val), -jnp.inf, x.log_abs_val))
    return x



from jaxns.framework.bases import BaseAbstractModel
from jaxns.nested_sampler.abc import AbstractNestedSampler


class BaseAbstractNestedSampler(AbstractNestedSampler):
    def __init__(self, model: BaseAbstractModel, max_samples: int):
        self.model = model
        self.max_samples = max_samples



from typing import NamedTuple, Tuple

from jax import random, numpy as jnp, lax, tree_map

from jaxns.framework.bases import BaseAbstractModel
from jaxns.internals.types import IntArray, StaticStandardNestedSamplerState, UType, MeasureType, \
    StaticStandardSampleCollection
from jaxns.internals.types import PRNGKey, FloatArray
from jaxns.internals.types import Sample, int_type
from jaxns.samplers.abc import SamplerState
from jaxns.samplers.bases import BaseAbstractRejectionSampler

__all__ = [
    'UniformSampler'
]


class UniformSampler(BaseAbstractRejectionSampler):
    """
    A sampler that produces uniform samples from the model within the log_L_constraint.
    """

    def __init__(self, model: BaseAbstractModel, max_likelihood_evals: int = 100):
        """
        Initialises the sampler.

        Args:
            model: the model to sample from
            max_likelihood_evals: the maximum number of likelihood evaluations to perform, before stopping. This is
                important for not getting stuck on plateaus, or forbidden zones.
        """
        super().__init__(model=model)
        if max_likelihood_evals < 1:
            raise ValueError("max_likelihood_evals must be >= 1")
        self.max_likelihood_evals = int(max_likelihood_evals)

    def num_phantom(self) -> int:
        return 0

    def pre_process(self, state: StaticStandardNestedSamplerState) -> SamplerState:
        return ()

    def post_process(self, sample_collection: StaticStandardSampleCollection,
                     sampler_state: SamplerState) -> SamplerState:
        return sampler_state

    def get_sample(self, key: PRNGKey, log_L_constraint: FloatArray, sampler_state: SamplerState) -> Tuple[
        Sample, Sample]:
        class CarryState(NamedTuple):
            key: PRNGKey
            U: UType
            log_L: MeasureType
            num_likelihood_evals: IntArray

        def cond(carry_state: CarryState):
            done_1 = carry_state.log_L > log_L_constraint
            done_2 = jnp.bitwise_or(
                carry_state.log_L == log_L_constraint,
                carry_state.num_likelihood_evals >= self.max_likelihood_evals
            )
            done = jnp.bitwise_or(done_1, done_2)
            return jnp.bitwise_not(done)

        def body(carry_state: CarryState) -> CarryState:
            key, sample_key = random.split(carry_state.key, 2)
            U = self.model.sample_U(key=sample_key)
            log_L = self.model.forward(U=U)
            num_likelihood_evals = carry_state.num_likelihood_evals + jnp.ones_like(carry_state.num_likelihood_evals)
            return CarryState(key=key, U=U, log_L=log_L, num_likelihood_evals=num_likelihood_evals)

        key, sample_key = random.split(key, 2)
        init_U = self.model.sample_U(key=sample_key)
        init_log_L = self.model.forward(init_U)
        init_carry_state = CarryState(
            key=key,
            U=init_U,
            log_L=init_log_L,
            num_likelihood_evals=jnp.asarray(1, int_type)
        )

        carry_state = lax.while_loop(cond_fun=cond, body_fun=body, init_val=init_carry_state)

        sample = Sample(
            U_sample=carry_state.U,
            log_L_constraint=log_L_constraint,
            log_L=carry_state.log_L,
            num_likelihood_evaluations=carry_state.num_likelihood_evals
        )
        phantom_samples = tree_map(lambda x: jnp.zeros((0,) + x.shape, x.dtype), sample)
        return sample, phantom_samples



import numpy as np
from jax import random, numpy as jnp

from jaxns.plotting import weighted_percentile
from jaxns.utils import resample, _bit_mask


def test_resample():
    x = random.normal(key=random.PRNGKey(0), shape=(50,))
    logits = -jnp.ones(50)
    samples = {'x': x}
    assert jnp.all(resample(random.PRNGKey(0), samples, logits)['x'] == resample(random.PRNGKey(0), x, logits))


def test_bit_mask():
    assert _bit_mask(1, width=2) == [1, 0]
    assert _bit_mask(2, width=2) == [0, 1]
    assert _bit_mask(3, width=2) == [1, 1]


def test_weighted_percentile():
    # Test the weighted percentile function
    samples = np.asarray([1, 2, 3, 4, 5])
    log_weights = np.asarray([0, 0, 0, 0, 0])
    percentiles = [50]
    assert np.allclose(weighted_percentile(samples, log_weights, percentiles), 3.0)



from jax import numpy as jnp

from jaxns.internals.cumulative_ops import cumulative_op_static, cumulative_op_dynamic
from jaxns.internals.types import float_type, int_type


def test_cumulative_op_static():
    def op(accumulate, y):
        return accumulate + y

    init = jnp.asarray(0, float_type)
    xs = jnp.asarray([1, 2, 3], float_type)
    final_accumulate, result = cumulative_op_static(op=op, init=init, xs=xs)
    assert final_accumulate == 6
    assert all(result == jnp.asarray([1, 3, 6], float_type))

    final_accumulate, result = cumulative_op_static(op=op, init=init, xs=xs, pre_op=True)
    assert final_accumulate == 6
    assert all(result == jnp.asarray([0, 1, 3], float_type))


def test_cumulative_op_dynamic():
    def op(accumulate, y):
        return accumulate + y

    init = jnp.asarray(0, float_type)
    xs = jnp.asarray([1, 2, 3], float_type)
    stop_idx = jnp.asarray(3, int_type)
    final_accumulate, result = cumulative_op_dynamic(op=op, init=init, xs=xs, stop_idx=stop_idx)
    assert final_accumulate == 6
    assert all(result == jnp.asarray([1, 3, 6], float_type))

    final_accumulate, result = cumulative_op_dynamic(op=op, init=init, xs=xs, stop_idx=stop_idx, pre_op=True)
    assert final_accumulate == 6
    assert all(result == jnp.asarray([0, 1, 3], float_type))

    stop_idx = jnp.asarray(2, int_type)
    final_accumulate, result = cumulative_op_dynamic(op=op, init=init, xs=xs, stop_idx=stop_idx)
    assert final_accumulate == 3
    assert all(result == jnp.asarray([1, 3, 0], float_type))

    final_accumulate, result = cumulative_op_dynamic(op=op, init=init, xs=xs, stop_idx=stop_idx, pre_op=True)
    assert final_accumulate == 3
    assert all(result == jnp.asarray([0, 1, 0], float_type))



import logging
from typing import Optional
from uuid import uuid4

import numpy as np
from jax import random, vmap, jit, numpy as jnp

try:
    import haiku as hk
except ImportError:
    print("You must `pip install dm-haiku` first.")
    raise

from jaxns.framework.bases import BaseAbstractModel, PriorModelType
from jaxns.framework.ops import transform, prepare_input, compute_log_prob_prior, compute_log_likelihood, parse_prior, \
    parse_joint, transform_parametrised
from jaxns.internals.types import PRNGKey, FloatArray, float_type, LikelihoodType, UType, XType, LikelihoodInputType

__all__ = [
    'Model'
]

logger = logging.getLogger('jaxns')


class Model(BaseAbstractModel):
    """
    Represents a Bayesian model in terms of a generative prior, and likelihood function.
    """

    def __init__(self, prior_model: PriorModelType, log_likelihood: LikelihoodType,
                 params: Optional[hk.MutableParams] = None):
        super().__init__(prior_model=prior_model, log_likelihood=log_likelihood)
        # hk.transform(
        #     lambda: parse_prior(prior_model=self.prior_model)
        # ).apply(params=self._params, rng=None)
        if params is None:
            params = self.init_params(rng=random.PRNGKey(0))
        self._params = params
        # Parse the prior model to get place holders
        self.__U_placeholder, self.__X_placeholder = hk.transform(
            lambda: parse_prior(prior_model=self.prior_model)
        ).apply(params=self._params, rng=None)
        self._id = str(uuid4())  # Used for making sure it's hashable, so it can be used as a key in a dict.

    @property
    def num_params(self) -> int:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")
        return hk.data_structures.tree_size(self._params)

    @property
    def params(self):
        if self._params is None:
            raise RuntimeError("Model has not been initialised")
        return self._params

    def set_params(self, params: hk.MutableParams) -> 'Model':
        """
        Create a new parametrised model with the given parameters.

        Args:
            params: The parameters to use.

        Returns:
            A model with set parameters.
        """
        return Model(prior_model=self.prior_model, log_likelihood=self.log_likelihood, params=params)

    def __call__(self, params: hk.MutableParams) -> 'Model':
        """
        Create a new parametrised model with the given parameters.

        **This is (and must be) a pure function.**

        Args:
            params: The parameters to use.

        Returns:
            A model with set parameters.
        """
        return self.set_params(params=params)

    def _U_placeholder(self) -> UType:
        return self.__U_placeholder

    def _X_placeholder(self) -> XType:
        return self.__X_placeholder

    def init_params(self, rng: PRNGKey) -> hk.MutableParams:
        """
        Initialise the parameters of the model.

        Args:
            rng: PRNGkey to initialise the parameters.

        Returns:
            The initialised parameters.
        """

        def _log_prob_joint():
            U_placeholder, X_placeholder, likelihood_input_placeholder, log_L_placeholder = parse_joint(
                prior_model=self.prior_model,
                log_likelihood=self.log_likelihood
            )
            log_prob_prior = compute_log_prob_prior(
                U=U_placeholder,
                prior_model=self.prior_model)
            log_prob_likelihood = compute_log_likelihood(
                U=U_placeholder,
                prior_model=self.prior_model,
                log_likelihood=self.log_likelihood,
                allow_nan=True
            )
            return log_prob_prior + log_prob_likelihood

        params = hk.transform(_log_prob_joint).init(rng)
        return params

    def __hash__(self):
        return hash(self._id)

    def __repr__(self):
        # TODO(Joshuaalbert): Pretty print the model
        return f""

    def sample_U(self, key: PRNGKey) -> FloatArray:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")

        def _sample_U():
            return random.uniform(key=hk.next_rng_key(), shape=(self.U_ndims,), dtype=float_type)

        return hk.transform(_sample_U).apply(params=self._params, rng=key)

    def transform(self, U: UType) -> XType:
        def _transform():
            return transform(U=U, prior_model=self.prior_model)

        return hk.transform(_transform).apply(params=self._params, rng=None)

    def transform_parametrised(self, U: UType) -> XType:
        def _transform():
            return transform_parametrised(U=U, prior_model=self.prior_model)

        return hk.transform(_transform).apply(params=self._params, rng=None)

    def forward(self, U: UType, allow_nan: bool = False) -> FloatArray:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")

        def _forward():
            return compute_log_likelihood(U=U, prior_model=self.prior_model, log_likelihood=self.log_likelihood,
                                          allow_nan=allow_nan)

        return hk.transform(_forward).apply(params=self._params, rng=None)

    def log_prob_prior(self, U: UType) -> FloatArray:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")

        def _log_prob_prior():
            return compute_log_prob_prior(U=U, prior_model=self.prior_model)

        return hk.transform(_log_prob_prior).apply(params=self._params, rng=None)

    def prepare_input(self, U: UType) -> LikelihoodInputType:
        if self._params is None:
            raise RuntimeError("Model has not been initialised")

        def _prepare_input():
            return prepare_input(U=U, prior_model=self.prior_model)

        return hk.transform(_prepare_input).apply(params=self._params, rng=None)

    def sanity_check(self, key: PRNGKey, S: int):
        U = jit(vmap(self.sample_U))(random.split(key, S))
        log_L = jit(vmap(lambda u: self.forward(u, allow_nan=True)))(U)
        logger.info("Sanity check...")
        for _U, _log_L in zip(U, log_L):
            if jnp.isnan(_log_L):
                logger.info(f"Found bad point: {_U} -> {self.transform(_U)}")
        assert not any(np.isnan(log_L))
        logger.info("Sanity check passed")
        if 'parsed_prior' in self.__dict__:
            del self.__dict__['parsed_prior']
